package com.itemis.jbc.binary;

import com.google.common.base.Objects;
import com.itemis.jbc.binary.ClassFileAccessAPI;
import com.itemis.jbc.binary.ClassFileFactoryAPI;
import com.itemis.jbc.binary.Opcode;
import com.itemis.jbc.jbc.AttributeInfo;
import com.itemis.jbc.jbc.Attributes;
import com.itemis.jbc.jbc.ClassFile;
import com.itemis.jbc.jbc.Code;
import com.itemis.jbc.jbc.CodeTable;
import com.itemis.jbc.jbc.CodeTableEntry;
import com.itemis.jbc.jbc.ConstantClass;
import com.itemis.jbc.jbc.ConstantDouble;
import com.itemis.jbc.jbc.ConstantFieldRef;
import com.itemis.jbc.jbc.ConstantFloat;
import com.itemis.jbc.jbc.ConstantInteger;
import com.itemis.jbc.jbc.ConstantInterfaceMethodRef;
import com.itemis.jbc.jbc.ConstantInvoceDynamic;
import com.itemis.jbc.jbc.ConstantLong;
import com.itemis.jbc.jbc.ConstantMethodHandle;
import com.itemis.jbc.jbc.ConstantMethodRef;
import com.itemis.jbc.jbc.ConstantMethodType;
import com.itemis.jbc.jbc.ConstantModule;
import com.itemis.jbc.jbc.ConstantNameAndType;
import com.itemis.jbc.jbc.ConstantPackage;
import com.itemis.jbc.jbc.ConstantPool;
import com.itemis.jbc.jbc.ConstantPoolEntry;
import com.itemis.jbc.jbc.ConstantString;
import com.itemis.jbc.jbc.ConstantUtf8;
import com.itemis.jbc.jbc.ConstantValue;
import com.itemis.jbc.jbc.EnclosingMethod;
import com.itemis.jbc.jbc.ExceptionTable;
import com.itemis.jbc.jbc.ExceptionTableEntry;
import com.itemis.jbc.jbc.Exports;
import com.itemis.jbc.jbc.ExportsTo;
import com.itemis.jbc.jbc.FieldInfo;
import com.itemis.jbc.jbc.Fields;
import com.itemis.jbc.jbc.GOTO;
import com.itemis.jbc.jbc.GOTO_W;
import com.itemis.jbc.jbc.IFEQ;
import com.itemis.jbc.jbc.IFGE;
import com.itemis.jbc.jbc.IFGT;
import com.itemis.jbc.jbc.IFLE;
import com.itemis.jbc.jbc.IFLT;
import com.itemis.jbc.jbc.IFNE;
import com.itemis.jbc.jbc.IFNONNULL;
import com.itemis.jbc.jbc.IFNULL;
import com.itemis.jbc.jbc.IF_ACMPEQ;
import com.itemis.jbc.jbc.IF_ACMPNE;
import com.itemis.jbc.jbc.IF_ICMPEQ;
import com.itemis.jbc.jbc.IF_ICMPGE;
import com.itemis.jbc.jbc.IF_ICMPGT;
import com.itemis.jbc.jbc.IF_ICMPLE;
import com.itemis.jbc.jbc.IF_ICMPLT;
import com.itemis.jbc.jbc.IF_ICMPNE;
import com.itemis.jbc.jbc.InnerClass;
import com.itemis.jbc.jbc.InnerClasses;
import com.itemis.jbc.jbc.Interface;
import com.itemis.jbc.jbc.Interfaces;
import com.itemis.jbc.jbc.JbcFactory;
import com.itemis.jbc.jbc.LineNumber;
import com.itemis.jbc.jbc.LineNumberTable;
import com.itemis.jbc.jbc.LocalVariable;
import com.itemis.jbc.jbc.LocalVariableTable;
import com.itemis.jbc.jbc.MethodInfo;
import com.itemis.jbc.jbc.Methods;
import com.itemis.jbc.jbc.Module;
import com.itemis.jbc.jbc.Opens;
import com.itemis.jbc.jbc.OpensTo;
import com.itemis.jbc.jbc.Provides;
import com.itemis.jbc.jbc.ProvidesWith;
import com.itemis.jbc.jbc.Requires;
import com.itemis.jbc.jbc.SourceFile;
import com.itemis.jbc.jbc.U1;
import com.itemis.jbc.jbc.U2;
import com.itemis.jbc.jbc.U4;
import com.itemis.jbc.jbc.UString;
import com.itemis.jbc.jbc.Unknown;
import com.itemis.jbc.jbc.Uses;
import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;
import org.eclipse.xtext.xbase.lib.CollectionExtensions;
import org.eclipse.xtext.xbase.lib.Exceptions;
import org.eclipse.xtext.xbase.lib.Pair;

@SuppressWarnings("all")
public class ByteCodeReader {
  public static ClassFile fromByteArray(final byte[] content) {
    ByteArrayInputStream _byteArrayInputStream = new ByteArrayInputStream(content);
    DataInputStream _dataInputStream = new DataInputStream(_byteArrayInputStream);
    return ByteCodeReader.fromDataInputStream(_dataInputStream);
  }
  
  public static ClassFile fromDataInputStream(final DataInputStream stream) {
    return new ByteCodeReader(stream).readClassFile();
  }
  
  private DataInputStream stream;
  
  private ClassFile classFile;
  
  private CodeTable currentCodeTable;
  
  ByteCodeReader(final DataInputStream stream) {
    this.stream = stream;
  }
  
  ClassFile readClassFile() {
    final ClassFile result = JbcFactory.eINSTANCE.createClassFile();
    this.classFile = result;
    result.setMagic(this.readU4());
    result.setMinorVersion(this.readU2());
    result.setMajorVersion(this.readU2());
    result.setConstantPoolCount(this.readU2());
    result.setConstantPool(this.readConstantPool(result.getConstantPoolCount()));
    result.setAccessFlags(this.readU2());
    result.setThisClass(this.readConstantClassRef());
    result.setSuperClass(this.readConstantClassRef());
    result.setInterfaceCount(this.readU2());
    result.setInterfaces(this.readInterfaces(result.getInterfaceCount()));
    result.setFieldsCount(this.readU2());
    result.setFields(this.readFields(result.getFieldsCount()));
    result.setMethodsCount(this.readU2());
    result.setMethods(this.readMethods(result.getMethodsCount()));
    result.setAttributesCount(this.readU2());
    result.setAttributes(this.readAttributes(result.getAttributesCount()));
    return result;
  }
  
  ConstantPool readConstantPool(final U2 constantPoolCount) {
    try {
      final ConstantPool result = JbcFactory.eINSTANCE.createConstantPool();
      final LinkedHashMap<ConstantPoolEntry, Pair<Integer, Integer>> map = new LinkedHashMap<ConstantPoolEntry, Pair<Integer, Integer>>();
      for (int i = 0; (i < (ClassFileAccessAPI.intValue(constantPoolCount) - 1)); i++) {
        {
          final U1 tag = this.readU1();
          ConstantPoolEntry constant = null;
          Class<? extends ConstantPoolEntry> _entryTypeForTag = ClassFileAccessAPI.entryTypeForTag(ClassFileAccessAPI.intValue(tag));
          boolean _matched = false;
          if (Objects.equal(_entryTypeForTag, ConstantUtf8.class)) {
            _matched=true;
            constant = this.readConstantUtf8(tag);
          }
          if (!_matched) {
            if (Objects.equal(_entryTypeForTag, ConstantInteger.class)) {
              _matched=true;
              constant = ClassFileFactoryAPI.constantInteger(tag, this.readU4());
            }
          }
          if (!_matched) {
            if (Objects.equal(_entryTypeForTag, ConstantFloat.class)) {
              _matched=true;
              constant = ClassFileFactoryAPI.constantFloat(tag, this.readU4());
            }
          }
          if (!_matched) {
            if (Objects.equal(_entryTypeForTag, ConstantLong.class)) {
              _matched=true;
              constant = ClassFileFactoryAPI.constantLong(tag, this.readU4(), this.readU4());
              i++;
            }
          }
          if (!_matched) {
            if (Objects.equal(_entryTypeForTag, ConstantDouble.class)) {
              _matched=true;
              constant = ClassFileFactoryAPI.constantDouble(tag, this.readU4(), this.readU4());
              i++;
            }
          }
          if (!_matched) {
            if (Objects.equal(_entryTypeForTag, ConstantClass.class)) {
              _matched=true;
              map.put(constant = ClassFileFactoryAPI.constantClass(tag, null), Pair.<Integer, Integer>of(Integer.valueOf(this.stream.readUnsignedShort()), null));
            }
          }
          if (!_matched) {
            if (Objects.equal(_entryTypeForTag, ConstantString.class)) {
              _matched=true;
              map.put(constant = ClassFileFactoryAPI.constantString(tag, null), Pair.<Integer, Integer>of(Integer.valueOf(this.stream.readUnsignedShort()), null));
            }
          }
          if (!_matched) {
            if (Objects.equal(_entryTypeForTag, ConstantFieldRef.class)) {
              _matched=true;
              map.put(constant = ClassFileFactoryAPI.constantFieldRef(tag, null, null), 
                Pair.<Integer, Integer>of(Integer.valueOf(this.stream.readUnsignedShort()), Integer.valueOf(this.stream.readUnsignedShort())));
            }
          }
          if (!_matched) {
            if (Objects.equal(_entryTypeForTag, ConstantMethodRef.class)) {
              _matched=true;
              map.put(constant = ClassFileFactoryAPI.constantMethodRef(tag, null, null), 
                Pair.<Integer, Integer>of(Integer.valueOf(this.stream.readUnsignedShort()), Integer.valueOf(this.stream.readUnsignedShort())));
            }
          }
          if (!_matched) {
            if (Objects.equal(_entryTypeForTag, ConstantInterfaceMethodRef.class)) {
              _matched=true;
              map.put(constant = ClassFileFactoryAPI.constantInterfaceMethodRef(tag, null, null), 
                Pair.<Integer, Integer>of(Integer.valueOf(this.stream.readUnsignedShort()), Integer.valueOf(this.stream.readUnsignedShort())));
            }
          }
          if (!_matched) {
            if (Objects.equal(_entryTypeForTag, ConstantNameAndType.class)) {
              _matched=true;
              map.put(constant = ClassFileFactoryAPI.constantNameAndType(tag, null, null), 
                Pair.<Integer, Integer>of(Integer.valueOf(this.stream.readUnsignedShort()), Integer.valueOf(this.stream.readUnsignedShort())));
            }
          }
          if (!_matched) {
            if (Objects.equal(_entryTypeForTag, ConstantMethodHandle.class)) {
              _matched=true;
              map.put(constant = ClassFileFactoryAPI.constantMethodHandle(tag, this.readU1(), null), Pair.<Integer, Integer>of(Integer.valueOf(this.stream.readUnsignedShort()), null));
            }
          }
          if (!_matched) {
            if (Objects.equal(_entryTypeForTag, ConstantMethodType.class)) {
              _matched=true;
              map.put(constant = ClassFileFactoryAPI.constantMethodType(tag, null), Pair.<Integer, Integer>of(Integer.valueOf(this.stream.readUnsignedShort()), null));
            }
          }
          if (!_matched) {
            if (Objects.equal(_entryTypeForTag, ConstantInvoceDynamic.class)) {
              _matched=true;
              map.put(constant = ClassFileFactoryAPI.constantInvoceDynamic(tag, this.readU2(), null), 
                Pair.<Integer, Integer>of(Integer.valueOf(this.stream.readUnsignedShort()), null));
            }
          }
          if (!_matched) {
            if (Objects.equal(_entryTypeForTag, ConstantModule.class)) {
              _matched=true;
              map.put(constant = ClassFileFactoryAPI.constantModule(tag, null), Pair.<Integer, Integer>of(Integer.valueOf(this.stream.readUnsignedShort()), null));
            }
          }
          if (!_matched) {
            if (Objects.equal(_entryTypeForTag, ConstantPackage.class)) {
              _matched=true;
              map.put(constant = ClassFileFactoryAPI.constantPackage(tag, null), Pair.<Integer, Integer>of(Integer.valueOf(this.stream.readUnsignedShort()), null));
            }
          }
          if (!_matched) {
            throw new RuntimeException(("Unknown tag: " + tag));
          }
          result.getCpInfo().add(constant);
        }
      }
      this.addFromMaps(result, map);
      return result;
    } catch (Throwable _e) {
      throw Exceptions.sneakyThrow(_e);
    }
  }
  
  private void addFromMaps(final ConstantPool pool, final Map<ConstantPoolEntry, Pair<Integer, Integer>> map) {
    Set<Map.Entry<ConstantPoolEntry, Pair<Integer, Integer>>> _entrySet = map.entrySet();
    for (final Map.Entry<ConstantPoolEntry, Pair<Integer, Integer>> e : _entrySet) {
      {
        final ConstantPoolEntry entry = e.getKey();
        boolean _matched = false;
        if (entry instanceof ConstantClass) {
          _matched=true;
          ((ConstantClass)entry).setNameIndex(ClassFileAccessAPI.getConstantUtf8(pool, (e.getValue().getKey()).intValue()));
        }
        if (!_matched) {
          if (entry instanceof ConstantString) {
            _matched=true;
            ((ConstantString)entry).setStringIndex(ClassFileAccessAPI.getConstantUtf8(pool, (e.getValue().getKey()).intValue()));
          }
        }
        if (!_matched) {
          if (entry instanceof ConstantFieldRef) {
            _matched=true;
            ((ConstantFieldRef)entry).setClassIndex(ClassFileAccessAPI.getConstantClass(pool, (e.getValue().getKey()).intValue()));
            ((ConstantFieldRef)entry).setNameAndTypeIndex(ClassFileAccessAPI.getConstantNameAndType(pool, (e.getValue().getValue()).intValue()));
          }
        }
        if (!_matched) {
          if (entry instanceof ConstantMethodRef) {
            _matched=true;
            ((ConstantMethodRef)entry).setClassIndex(ClassFileAccessAPI.getConstantClass(pool, (e.getValue().getKey()).intValue()));
            ((ConstantMethodRef)entry).setNameAndTypeIndex(ClassFileAccessAPI.getConstantNameAndType(pool, (e.getValue().getValue()).intValue()));
          }
        }
        if (!_matched) {
          if (entry instanceof ConstantInterfaceMethodRef) {
            _matched=true;
            ((ConstantInterfaceMethodRef)entry).setClassIndex(ClassFileAccessAPI.getConstantClass(pool, (e.getValue().getKey()).intValue()));
            ((ConstantInterfaceMethodRef)entry).setNameAndTypeIndex(ClassFileAccessAPI.getConstantNameAndType(pool, (e.getValue().getValue()).intValue()));
          }
        }
        if (!_matched) {
          if (entry instanceof ConstantNameAndType) {
            _matched=true;
            ((ConstantNameAndType)entry).setNameIndex(ClassFileAccessAPI.getConstantUtf8(pool, (e.getValue().getKey()).intValue()));
            ((ConstantNameAndType)entry).setDescriptorIndex(ClassFileAccessAPI.getConstantUtf8(pool, (e.getValue().getValue()).intValue()));
          }
        }
        if (!_matched) {
          if (entry instanceof ConstantMethodHandle) {
            _matched=true;
            ((ConstantMethodHandle)entry).setReferenceIndex(ClassFileAccessAPI.getConstant(pool, (e.getValue().getKey()).intValue()));
          }
        }
        if (!_matched) {
          if (entry instanceof ConstantMethodType) {
            _matched=true;
            ((ConstantMethodType)entry).setDescriptorIndex(ClassFileAccessAPI.getConstantUtf8(pool, (e.getValue().getKey()).intValue()));
          }
        }
        if (!_matched) {
          if (entry instanceof ConstantInvoceDynamic) {
            _matched=true;
            ((ConstantInvoceDynamic)entry).setNameAndTypeIndex(ClassFileAccessAPI.getConstantNameAndType(pool, (e.getValue().getKey()).intValue()));
          }
        }
        if (!_matched) {
          if (entry instanceof ConstantModule) {
            _matched=true;
            ((ConstantModule)entry).setNameIndex(ClassFileAccessAPI.getConstantUtf8(pool, (e.getValue().getKey()).intValue()));
          }
        }
        if (!_matched) {
          if (entry instanceof ConstantPackage) {
            _matched=true;
            ((ConstantPackage)entry).setNameIndex(ClassFileAccessAPI.getConstantUtf8(pool, (e.getValue().getKey()).intValue()));
          }
        }
      }
    }
  }
  
  private ConstantUtf8 readConstantUtf8(final U1 tag) {
    final ConstantUtf8 result = JbcFactory.eINSTANCE.createConstantUtf8();
    result.setTag(tag);
    result.setContent(this.readString());
    return result;
  }
  
  private Interfaces readInterfaces(final U2 interfaceCount) {
    final Interfaces result = JbcFactory.eINSTANCE.createInterfaces();
    for (int i = 0; (i < ClassFileAccessAPI.intValue(interfaceCount)); i++) {
      result.getInterfaceInfo().add(this.readInterface());
    }
    return result;
  }
  
  private Interface readInterface() {
    final Interface result = JbcFactory.eINSTANCE.createInterface();
    result.setInfo(this.readConstantClassRef());
    return result;
  }
  
  private Fields readFields(final U2 fieldsCount) {
    final Fields result = JbcFactory.eINSTANCE.createFields();
    for (int i = 0; (i < ClassFileAccessAPI.intValue(fieldsCount)); i++) {
      result.getFieldInfo().add(this.readField());
    }
    return result;
  }
  
  private FieldInfo readField() {
    final FieldInfo result = JbcFactory.eINSTANCE.createFieldInfo();
    result.setAccessFlags(this.readU2());
    result.setNameIndex(this.readConstantUtf8Ref());
    result.setDescriptorIndex(this.readConstantUtf8Ref());
    result.setAttributesCount(this.readU2());
    result.setAttributes(this.readAttributes(result.getAttributesCount()));
    return result;
  }
  
  private Methods readMethods(final U2 methodsCount) {
    final Methods result = JbcFactory.eINSTANCE.createMethods();
    for (int i = 0; (i < ClassFileAccessAPI.intValue(methodsCount)); i++) {
      result.getMethodsInfo().add(this.readMethod());
    }
    return result;
  }
  
  private MethodInfo readMethod() {
    final MethodInfo result = JbcFactory.eINSTANCE.createMethodInfo();
    result.setAccessFlags(this.readU2());
    result.setNameIndex(this.readConstantUtf8Ref());
    result.setDescriptorIndex(this.readConstantUtf8Ref());
    result.setAttributesCount(this.readU2());
    result.setAttributes(this.readAttributes(result.getAttributesCount()));
    return result;
  }
  
  private Attributes readAttributes(final U2 attributesCount) {
    final Attributes result = JbcFactory.eINSTANCE.createAttributes();
    for (int i = 0; (i < ClassFileAccessAPI.intValue(attributesCount)); i++) {
      result.getAttributeInfo().add(this.readAttribute());
    }
    return result;
  }
  
  private AttributeInfo readAttribute() {
    final ConstantUtf8 attributeNameIndex = this.readConstantUtf8Ref();
    final U4 attributeLength = this.readU4();
    if ((attributeNameIndex == null)) {
      throw new RuntimeException();
    }
    boolean _equals = "ConstantValue".equals(ClassFileAccessAPI.getStringValue(attributeNameIndex));
    if (_equals) {
      return this.readConstantValueAttribute(attributeNameIndex, attributeLength);
    } else {
      boolean _equals_1 = "Code".equals(ClassFileAccessAPI.getStringValue(attributeNameIndex));
      if (_equals_1) {
        return this.readCodeAttribute(attributeNameIndex, attributeLength);
      } else {
        boolean _equals_2 = "SourceFile".equals(ClassFileAccessAPI.getStringValue(attributeNameIndex));
        if (_equals_2) {
          return this.readSourceFileAttribute(attributeNameIndex, attributeLength);
        } else {
          boolean _equals_3 = "LineNumberTable".equals(ClassFileAccessAPI.getStringValue(attributeNameIndex));
          if (_equals_3) {
            return this.readLineNumberTableAttribute(attributeNameIndex, attributeLength);
          } else {
            boolean _equals_4 = "LocalVariableTable".equals(ClassFileAccessAPI.getStringValue(attributeNameIndex));
            if (_equals_4) {
              return this.readLocalVariableTableAttribute(attributeNameIndex, attributeLength);
            } else {
              boolean _equals_5 = "Exceptions".equals(ClassFileAccessAPI.getStringValue(attributeNameIndex));
              if (_equals_5) {
                return this.readExceptionsAttribute(attributeNameIndex, attributeLength);
              } else {
                boolean _equals_6 = "InnerClasses".equals(ClassFileAccessAPI.getStringValue(attributeNameIndex));
                if (_equals_6) {
                  return this.readInnerClassesAttribute(attributeNameIndex, attributeLength);
                } else {
                  boolean _equals_7 = "EnclosingMethod".equals(ClassFileAccessAPI.getStringValue(attributeNameIndex));
                  if (_equals_7) {
                    return this.readEnclosingMethodAttribute(attributeNameIndex, attributeLength);
                  } else {
                    boolean _equals_8 = "Module".equals(ClassFileAccessAPI.getStringValue(attributeNameIndex));
                    if (_equals_8) {
                      return this.readModuleAttribute(attributeNameIndex, attributeLength);
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
    return this.readUnknownAttribute(attributeNameIndex, attributeLength);
  }
  
  public Unknown readUnknownAttribute(final ConstantUtf8 attributeNameIndex, final U4 attributeLength) {
    final Unknown result = JbcFactory.eINSTANCE.createUnknown();
    result.setAttributeNameIndex(attributeNameIndex);
    result.setAttributeLength(attributeLength);
    CollectionExtensions.<U1>addAll(result.getInfo(), this.readBytes(ClassFileAccessAPI.intValue(result.getAttributeLength())));
    return result;
  }
  
  public ConstantValue readConstantValueAttribute(final ConstantUtf8 attributeNameIndex, final U4 attributeLength) {
    final ConstantValue result = JbcFactory.eINSTANCE.createConstantValue();
    result.setAttributeNameIndex(attributeNameIndex);
    result.setAttributeLength(attributeLength);
    result.setConstantValueIndex(this.readConstantPoolEntryRef());
    return result;
  }
  
  public Code readCodeAttribute(final ConstantUtf8 attributeNameIndex, final U4 attributeLength) {
    final Code result = JbcFactory.eINSTANCE.createCode();
    result.setAttributeNameIndex(attributeNameIndex);
    result.setAttributeLength(attributeLength);
    result.setMaxStack(this.readU2());
    result.setMaxLocals(this.readU2());
    result.setCodeLength(this.readU4());
    result.setCodeTable(this.readCodeTable(result.getCodeLength()));
    result.setExceptionTableLength(this.readU2());
    result.setExceptionTable(this.readExceptionTable(result.getExceptionTableLength()));
    result.setAttributesCount(this.readU2());
    result.setAttributes(this.readAttributes(result.getAttributesCount()));
    return result;
  }
  
  private ExceptionTable readExceptionTable(final U2 exceptionTableLength) {
    final ExceptionTable result = JbcFactory.eINSTANCE.createExceptionTable();
    for (int i = 0; (i < ClassFileAccessAPI.intValue(exceptionTableLength)); i++) {
      result.getExceptionTableEntry().add(this.readExceptionTableEntry());
    }
    return result;
  }
  
  private ExceptionTableEntry readExceptionTableEntry() {
    final ExceptionTableEntry result = JbcFactory.eINSTANCE.createExceptionTableEntry();
    result.setStartPc(this.readCodeTableEntryRef());
    result.setEndPc(this.readCodeTableEntryRef());
    result.setHandlerPc(this.readCodeTableEntryRef());
    result.setCatchType(this.readConstantClassRef());
    return result;
  }
  
  private CodeTable readCodeTable(final U4 codeLength) {
    final JbcFactory f = JbcFactory.eINSTANCE;
    final CodeTable result = f.createCodeTable();
    this.currentCodeTable = result;
    final Map<CodeTableEntry, Pair<Integer, Integer>> map = new LinkedHashMap<CodeTableEntry, Pair<Integer, Integer>>();
    final int length = ClassFileAccessAPI.intValue(codeLength);
    int offset = 0;
    while ((offset < length)) {
      {
        final U1 tag = this.readU1();
        final Opcode opcode = Opcode.from(ClassFileAccessAPI.intValue(tag));
        final CodeTableEntry instance = opcode.readInstance(this, offset, map);
        instance.setTag(tag);
        int _offset = offset;
        int _byteCount = this.getByteCount(instance);
        offset = (_offset + _byteCount);
        result.getInstruction().add(instance);
      }
    }
    Set<Map.Entry<CodeTableEntry, Pair<Integer, Integer>>> _entrySet = map.entrySet();
    for (final Map.Entry<CodeTableEntry, Pair<Integer, Integer>> entry : _entrySet) {
      {
        final CodeTableEntry code = entry.getKey();
        boolean _matched = false;
        if (code instanceof GOTO) {
          _matched=true;
          CodeTable _table = ClassFileAccessAPI.getTable(code);
          int _offset = ClassFileAccessAPI.offset(code);
          Integer _key = entry.getValue().getKey();
          int _plus = (_offset + (_key).intValue());
          ((GOTO)code).setBranch(ClassFileAccessAPI.instructionFromOffset(_table, _plus));
        }
        if (!_matched) {
          if (code instanceof GOTO_W) {
            _matched=true;
            CodeTable _table = ClassFileAccessAPI.getTable(code);
            int _offset = ClassFileAccessAPI.offset(code);
            Integer _key = entry.getValue().getKey();
            int _plus = (_offset + (_key).intValue());
            ((GOTO_W)code).setBranch(ClassFileAccessAPI.instructionFromOffset(_table, _plus));
          }
        }
        if (!_matched) {
          if (code instanceof IF_ACMPEQ) {
            _matched=true;
            CodeTable _table = ClassFileAccessAPI.getTable(code);
            int _offset = ClassFileAccessAPI.offset(code);
            Integer _key = entry.getValue().getKey();
            int _plus = (_offset + (_key).intValue());
            ((IF_ACMPEQ)code).setBranch(ClassFileAccessAPI.instructionFromOffset(_table, _plus));
          }
        }
        if (!_matched) {
          if (code instanceof IF_ACMPNE) {
            _matched=true;
            CodeTable _table = ClassFileAccessAPI.getTable(code);
            int _offset = ClassFileAccessAPI.offset(code);
            Integer _key = entry.getValue().getKey();
            int _plus = (_offset + (_key).intValue());
            ((IF_ACMPNE)code).setBranch(ClassFileAccessAPI.instructionFromOffset(_table, _plus));
          }
        }
        if (!_matched) {
          if (code instanceof IF_ICMPEQ) {
            _matched=true;
            CodeTable _table = ClassFileAccessAPI.getTable(code);
            int _offset = ClassFileAccessAPI.offset(code);
            Integer _key = entry.getValue().getKey();
            int _plus = (_offset + (_key).intValue());
            ((IF_ICMPEQ)code).setBranch(ClassFileAccessAPI.instructionFromOffset(_table, _plus));
          }
        }
        if (!_matched) {
          if (code instanceof IF_ICMPNE) {
            _matched=true;
            CodeTable _table = ClassFileAccessAPI.getTable(code);
            int _offset = ClassFileAccessAPI.offset(code);
            Integer _key = entry.getValue().getKey();
            int _plus = (_offset + (_key).intValue());
            ((IF_ICMPNE)code).setBranch(ClassFileAccessAPI.instructionFromOffset(_table, _plus));
          }
        }
        if (!_matched) {
          if (code instanceof IF_ICMPLT) {
            _matched=true;
            CodeTable _table = ClassFileAccessAPI.getTable(code);
            int _offset = ClassFileAccessAPI.offset(code);
            Integer _key = entry.getValue().getKey();
            int _plus = (_offset + (_key).intValue());
            ((IF_ICMPLT)code).setBranch(ClassFileAccessAPI.instructionFromOffset(_table, _plus));
          }
        }
        if (!_matched) {
          if (code instanceof IF_ICMPGE) {
            _matched=true;
            CodeTable _table = ClassFileAccessAPI.getTable(code);
            int _offset = ClassFileAccessAPI.offset(code);
            Integer _key = entry.getValue().getKey();
            int _plus = (_offset + (_key).intValue());
            ((IF_ICMPGE)code).setBranch(ClassFileAccessAPI.instructionFromOffset(_table, _plus));
          }
        }
        if (!_matched) {
          if (code instanceof IF_ICMPGT) {
            _matched=true;
            CodeTable _table = ClassFileAccessAPI.getTable(code);
            int _offset = ClassFileAccessAPI.offset(code);
            Integer _key = entry.getValue().getKey();
            int _plus = (_offset + (_key).intValue());
            ((IF_ICMPGT)code).setBranch(ClassFileAccessAPI.instructionFromOffset(_table, _plus));
          }
        }
        if (!_matched) {
          if (code instanceof IF_ICMPLE) {
            _matched=true;
            CodeTable _table = ClassFileAccessAPI.getTable(code);
            int _offset = ClassFileAccessAPI.offset(code);
            Integer _key = entry.getValue().getKey();
            int _plus = (_offset + (_key).intValue());
            ((IF_ICMPLE)code).setBranch(ClassFileAccessAPI.instructionFromOffset(_table, _plus));
          }
        }
        if (!_matched) {
          if (code instanceof IFEQ) {
            _matched=true;
            CodeTable _table = ClassFileAccessAPI.getTable(code);
            int _offset = ClassFileAccessAPI.offset(code);
            Integer _key = entry.getValue().getKey();
            int _plus = (_offset + (_key).intValue());
            ((IFEQ)code).setBranch(ClassFileAccessAPI.instructionFromOffset(_table, _plus));
          }
        }
        if (!_matched) {
          if (code instanceof IFNE) {
            _matched=true;
            CodeTable _table = ClassFileAccessAPI.getTable(code);
            int _offset = ClassFileAccessAPI.offset(code);
            Integer _key = entry.getValue().getKey();
            int _plus = (_offset + (_key).intValue());
            ((IFNE)code).setBranch(ClassFileAccessAPI.instructionFromOffset(_table, _plus));
          }
        }
        if (!_matched) {
          if (code instanceof IFLT) {
            _matched=true;
            CodeTable _table = ClassFileAccessAPI.getTable(code);
            int _offset = ClassFileAccessAPI.offset(code);
            Integer _key = entry.getValue().getKey();
            int _plus = (_offset + (_key).intValue());
            ((IFLT)code).setBranch(ClassFileAccessAPI.instructionFromOffset(_table, _plus));
          }
        }
        if (!_matched) {
          if (code instanceof IFGE) {
            _matched=true;
            CodeTable _table = ClassFileAccessAPI.getTable(code);
            int _offset = ClassFileAccessAPI.offset(code);
            Integer _key = entry.getValue().getKey();
            int _plus = (_offset + (_key).intValue());
            ((IFGE)code).setBranch(ClassFileAccessAPI.instructionFromOffset(_table, _plus));
          }
        }
        if (!_matched) {
          if (code instanceof IFGT) {
            _matched=true;
            CodeTable _table = ClassFileAccessAPI.getTable(code);
            int _offset = ClassFileAccessAPI.offset(code);
            Integer _key = entry.getValue().getKey();
            int _plus = (_offset + (_key).intValue());
            ((IFGT)code).setBranch(ClassFileAccessAPI.instructionFromOffset(_table, _plus));
          }
        }
        if (!_matched) {
          if (code instanceof IFLE) {
            _matched=true;
            CodeTable _table = ClassFileAccessAPI.getTable(code);
            int _offset = ClassFileAccessAPI.offset(code);
            Integer _key = entry.getValue().getKey();
            int _plus = (_offset + (_key).intValue());
            ((IFLE)code).setBranch(ClassFileAccessAPI.instructionFromOffset(_table, _plus));
          }
        }
        if (!_matched) {
          if (code instanceof IFNONNULL) {
            _matched=true;
            CodeTable _table = ClassFileAccessAPI.getTable(code);
            int _offset = ClassFileAccessAPI.offset(code);
            Integer _key = entry.getValue().getKey();
            int _plus = (_offset + (_key).intValue());
            ((IFNONNULL)code).setBranch(ClassFileAccessAPI.instructionFromOffset(_table, _plus));
          }
        }
        if (!_matched) {
          if (code instanceof IFNULL) {
            _matched=true;
            CodeTable _table = ClassFileAccessAPI.getTable(code);
            int _offset = ClassFileAccessAPI.offset(code);
            Integer _key = entry.getValue().getKey();
            int _plus = (_offset + (_key).intValue());
            ((IFNULL)code).setBranch(ClassFileAccessAPI.instructionFromOffset(_table, _plus));
          }
        }
        if (!_matched) {
          throw new RuntimeException();
        }
      }
    }
    return result;
  }
  
  private SourceFile readSourceFileAttribute(final ConstantUtf8 attributeNameIndex, final U4 attributeLength) {
    final SourceFile result = JbcFactory.eINSTANCE.createSourceFile();
    result.setAttributeNameIndex(attributeNameIndex);
    result.setAttributeLength(attributeLength);
    result.setSourceFileIndex(this.readConstantUtf8Ref());
    return result;
  }
  
  private LineNumberTable readLineNumberTableAttribute(final ConstantUtf8 attributeNameIndex, final U4 attributeLength) {
    final LineNumberTable result = JbcFactory.eINSTANCE.createLineNumberTable();
    result.setAttributeNameIndex(attributeNameIndex);
    result.setAttributeLength(attributeLength);
    result.setLineNumberTableLength(this.readU2());
    for (int n = 0; (n < ClassFileAccessAPI.intValue(result.getLineNumberTableLength())); n++) {
      result.getLineNumbers().add(this.readLineNumber());
    }
    return result;
  }
  
  private LineNumber readLineNumber() {
    final LineNumber result = JbcFactory.eINSTANCE.createLineNumber();
    result.setStartPc(this.readCodeTableEntryRef());
    result.setLineNumber(this.readU2());
    return result;
  }
  
  private LocalVariableTable readLocalVariableTableAttribute(final ConstantUtf8 attributeNameIndex, final U4 attributeLength) {
    final LocalVariableTable result = JbcFactory.eINSTANCE.createLocalVariableTable();
    result.setAttributeNameIndex(attributeNameIndex);
    result.setAttributeLength(attributeLength);
    result.setLocalVariableTableLength(this.readU2());
    for (int n = 0; (n < ClassFileAccessAPI.intValue(result.getLocalVariableTableLength())); n++) {
      result.getLocalVariables().add(this.readLocalVariable());
    }
    return result;
  }
  
  private LocalVariable readLocalVariable() {
    final LocalVariable result = JbcFactory.eINSTANCE.createLocalVariable();
    result.setStartPc(this.readCodeTableEntryRef());
    result.setLength(this.readU2());
    result.setNameIndex(this.readConstantUtf8Ref());
    result.setDescriptorIndex(this.readConstantUtf8Ref());
    result.setIndex(this.readU2());
    return result;
  }
  
  private com.itemis.jbc.jbc.Exceptions readExceptionsAttribute(final ConstantUtf8 attributeNameIndex, final U4 attributeLength) {
    final com.itemis.jbc.jbc.Exceptions result = JbcFactory.eINSTANCE.createExceptions();
    result.setAttributeNameIndex(attributeNameIndex);
    result.setAttributeLength(attributeLength);
    result.setNumberOfExceptions(this.readU2());
    for (int n = 0; (n < ClassFileAccessAPI.intValue(result.getNumberOfExceptions())); n++) {
      result.getException().add(this.readException());
    }
    return result;
  }
  
  private com.itemis.jbc.jbc.Exception readException() {
    final com.itemis.jbc.jbc.Exception result = JbcFactory.eINSTANCE.createException();
    result.setIndex(this.readConstantClassRef());
    return result;
  }
  
  private InnerClasses readInnerClassesAttribute(final ConstantUtf8 attributeNameIndex, final U4 attributeLength) {
    final InnerClasses result = JbcFactory.eINSTANCE.createInnerClasses();
    result.setAttributeNameIndex(attributeNameIndex);
    result.setAttributeLength(attributeLength);
    result.setNumberOfClasses(this.readU2());
    for (int n = 0; (n < ClassFileAccessAPI.intValue(result.getNumberOfClasses())); n++) {
      result.getInnerClasses().add(this.readInnerClass());
    }
    return result;
  }
  
  private InnerClass readInnerClass() {
    final InnerClass result = JbcFactory.eINSTANCE.createInnerClass();
    result.setInnerClassInfoIndex(this.readConstantClassRef());
    result.setOuterClassInfoIndex(this.readConstantClassRef());
    result.setInnerNameIndex(this.readConstantUtf8Ref());
    result.setInnerClassAccessFlags(this.readU2());
    return result;
  }
  
  private EnclosingMethod readEnclosingMethodAttribute(final ConstantUtf8 attributeNameIndex, final U4 attributeLength) {
    final EnclosingMethod result = JbcFactory.eINSTANCE.createEnclosingMethod();
    result.setAttributeNameIndex(attributeNameIndex);
    result.setAttributeLength(attributeLength);
    result.setClassIndex(this.readConstantClassRef());
    result.setMethodIndex(this.readConstantNameAndTypeRef());
    return result;
  }
  
  private Module readModuleAttribute(final ConstantUtf8 attributeNameIndex, final U4 attributeLength) {
    final Module result = JbcFactory.eINSTANCE.createModule();
    result.setAttributeNameIndex(attributeNameIndex);
    result.setAttributeLength(attributeLength);
    result.setModuleNameIndex(this.readConstantModuleRef());
    result.setModuleFlags(this.readU2());
    result.setModuleVersionIndex(this.readConstantUtf8Ref());
    result.setRequiresCount(this.readU2());
    for (int n = 0; (n < ClassFileAccessAPI.intValue(result.getRequiresCount())); n++) {
      result.getRequires().add(this.readRequires());
    }
    result.setExportsCount(this.readU2());
    for (int n = 0; (n < ClassFileAccessAPI.intValue(result.getExportsCount())); n++) {
      result.getExports().add(this.readExports());
    }
    result.setOpensCount(this.readU2());
    for (int n = 0; (n < ClassFileAccessAPI.intValue(result.getOpensCount())); n++) {
      result.getOpens().add(this.readOpens());
    }
    result.setUsesCount(this.readU2());
    for (int n = 0; (n < ClassFileAccessAPI.intValue(result.getUsesCount())); n++) {
      result.getUses().add(this.readUses());
    }
    result.setProvidesCount(this.readU2());
    for (int n = 0; (n < ClassFileAccessAPI.intValue(result.getProvidesCount())); n++) {
      result.getProvides().add(this.readProvides());
    }
    return result;
  }
  
  Requires readRequires() {
    final Requires result = JbcFactory.eINSTANCE.createRequires();
    result.setRequiresIndex(this.readConstantModuleRef());
    result.setRequiresFlags(this.readU2());
    result.setRequiresVersionIndex(this.readConstantUtf8Ref());
    return result;
  }
  
  Exports readExports() {
    final Exports result = JbcFactory.eINSTANCE.createExports();
    result.setExportsIndex(this.readConstantPackageRef());
    result.setExportsFlags(this.readU2());
    result.setExportsToCount(this.readU2());
    for (int n = 0; (n < ClassFileAccessAPI.intValue(result.getExportsToCount())); n++) {
      result.getExportsTo().add(this.readExportsTo());
    }
    return result;
  }
  
  ExportsTo readExportsTo() {
    final ExportsTo result = JbcFactory.eINSTANCE.createExportsTo();
    result.setExportsToIndex(this.readConstantModuleRef());
    return result;
  }
  
  Opens readOpens() {
    final Opens result = JbcFactory.eINSTANCE.createOpens();
    result.setOpensIndex(this.readConstantPackageRef());
    result.setOpensFlags(this.readU2());
    result.setOpensToCount(this.readU2());
    for (int n = 0; (n < ClassFileAccessAPI.intValue(result.getOpensToCount())); n++) {
      result.getOpensTo().add(this.readOpensTo());
    }
    return result;
  }
  
  OpensTo readOpensTo() {
    final OpensTo result = JbcFactory.eINSTANCE.createOpensTo();
    result.setOpensToIndex(this.readConstantModuleRef());
    return result;
  }
  
  Uses readUses() {
    final Uses result = JbcFactory.eINSTANCE.createUses();
    result.setUsesIndex(this.readConstantClassRef());
    return result;
  }
  
  Provides readProvides() {
    final Provides result = JbcFactory.eINSTANCE.createProvides();
    result.setProvidesIndex(this.readConstantClassRef());
    result.setProvidesWithCount(this.readU2());
    for (int n = 0; (n < ClassFileAccessAPI.intValue(result.getProvidesWithCount())); n++) {
      result.getProvidesWith().add(this.readProvidesWith());
    }
    return result;
  }
  
  ProvidesWith readProvidesWith() {
    final ProvidesWith result = JbcFactory.eINSTANCE.createProvidesWith();
    result.setProvidesWithIndex(this.readConstantClassRef());
    return result;
  }
  
  ConstantPoolEntry readConstantPoolEntryRef() {
    try {
      return ClassFileAccessAPI.getConstant(this.classFile.getConstantPool(), this.stream.readUnsignedShort());
    } catch (Throwable _e) {
      throw Exceptions.sneakyThrow(_e);
    }
  }
  
  ConstantPoolEntry readConstantPoolEntryRefFromByte() {
    try {
      return ClassFileAccessAPI.getConstant(this.classFile.getConstantPool(), this.stream.readUnsignedByte());
    } catch (Throwable _e) {
      throw Exceptions.sneakyThrow(_e);
    }
  }
  
  ConstantUtf8 readConstantUtf8Ref() {
    try {
      return ClassFileAccessAPI.getConstantUtf8(this.classFile.getConstantPool(), this.stream.readUnsignedShort());
    } catch (Throwable _e) {
      throw Exceptions.sneakyThrow(_e);
    }
  }
  
  ConstantClass readConstantClassRef() {
    try {
      return ClassFileAccessAPI.getConstantClass(this.classFile.getConstantPool(), this.stream.readUnsignedShort());
    } catch (Throwable _e) {
      throw Exceptions.sneakyThrow(_e);
    }
  }
  
  ConstantFieldRef readConstantFieldRef() {
    try {
      return ClassFileAccessAPI.getConstantFieldRef(this.classFile.getConstantPool(), this.stream.readUnsignedShort());
    } catch (Throwable _e) {
      throw Exceptions.sneakyThrow(_e);
    }
  }
  
  ConstantMethodRef readConstantMethodRef() {
    try {
      return ClassFileAccessAPI.getConstantMethodRef(this.classFile.getConstantPool(), this.stream.readUnsignedShort());
    } catch (Throwable _e) {
      throw Exceptions.sneakyThrow(_e);
    }
  }
  
  ConstantNameAndType readConstantNameAndTypeRef() {
    try {
      return ClassFileAccessAPI.getConstantNameAndType(this.classFile.getConstantPool(), this.stream.readUnsignedShort());
    } catch (Throwable _e) {
      throw Exceptions.sneakyThrow(_e);
    }
  }
  
  ConstantModule readConstantModuleRef() {
    try {
      return ClassFileAccessAPI.getConstantModule(this.classFile.getConstantPool(), this.stream.readUnsignedShort());
    } catch (Throwable _e) {
      throw Exceptions.sneakyThrow(_e);
    }
  }
  
  ConstantPackage readConstantPackageRef() {
    try {
      return ClassFileAccessAPI.getConstantPackage(this.classFile.getConstantPool(), this.stream.readUnsignedShort());
    } catch (Throwable _e) {
      throw Exceptions.sneakyThrow(_e);
    }
  }
  
  CodeTableEntry readCodeTableEntryRef() {
    try {
      return ClassFileAccessAPI.instructionFromOffset(this.currentCodeTable, this.stream.readUnsignedShort());
    } catch (Throwable _e) {
      throw Exceptions.sneakyThrow(_e);
    }
  }
  
  U1 readU1() {
    try {
      final U1 result = JbcFactory.eINSTANCE.createU1();
      result.setValue(String.format("%02X", Integer.valueOf(this.stream.readUnsignedByte())));
      return result;
    } catch (Throwable _e) {
      throw Exceptions.sneakyThrow(_e);
    }
  }
  
  U2 readU2() {
    try {
      final U2 result = JbcFactory.eINSTANCE.createU2();
      result.setValue(String.format("%04X", Integer.valueOf(this.stream.readUnsignedShort())));
      return result;
    } catch (Throwable _e) {
      throw Exceptions.sneakyThrow(_e);
    }
  }
  
  U4 readU4() {
    try {
      final U4 result = JbcFactory.eINSTANCE.createU4();
      result.setValue(String.format("%08X", Integer.valueOf(this.stream.readInt())));
      return result;
    } catch (Throwable _e) {
      throw Exceptions.sneakyThrow(_e);
    }
  }
  
  UString readString() {
    try {
      final UString result = JbcFactory.eINSTANCE.createUString();
      result.setValue(this.stream.readUTF());
      return result;
    } catch (Throwable _e) {
      throw Exceptions.sneakyThrow(_e);
    }
  }
  
  U1[] readBytes(final int count) {
    final U1[] result = new U1[count];
    for (int i = 0; (i < count); i++) {
      result[i] = this.readU1();
    }
    return result;
  }
  
  int readUnsignedShort() {
    try {
      return this.stream.readUnsignedShort();
    } catch (Throwable _e) {
      throw Exceptions.sneakyThrow(_e);
    }
  }
  
  int readShort() {
    try {
      return this.stream.readShort();
    } catch (Throwable _e) {
      throw Exceptions.sneakyThrow(_e);
    }
  }
  
  int readInt() {
    try {
      return this.stream.readInt();
    } catch (Throwable _e) {
      throw Exceptions.sneakyThrow(_e);
    }
  }
  
  private int getByteCount(final CodeTableEntry entry) {
    return Opcode.byteCount(entry);
  }
}
