package com.itemis.jbc.binary;

import com.itemis.jbc.binary.ClassFileAccessAPI;
import com.itemis.jbc.jbc.ClassFile;
import com.itemis.jbc.jbc.CodeTableEntry;
import com.itemis.jbc.jbc.ConstantPoolEntry;
import com.itemis.jbc.jbc.JbcPackage;
import com.itemis.jbc.jbc.U1;
import com.itemis.jbc.jbc.U2;
import com.itemis.jbc.jbc.U4;
import com.itemis.jbc.jbc.UString;
import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.util.List;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.xtext.xbase.lib.Exceptions;

@SuppressWarnings("all")
public class ByteCodeWriter {
  public static byte[] writeClassFile(final ClassFile classFile) {
    return new ByteCodeWriter(classFile).toByteArray();
  }
  
  private ClassFile classFile;
  
  ByteCodeWriter(final ClassFile classFile) {
    this.classFile = classFile;
  }
  
  private byte[] toByteArray() {
    final ByteArrayOutputStream byteArrayOutputStream = new ByteArrayOutputStream();
    DataOutputStream _dataOutputStream = new DataOutputStream(byteArrayOutputStream);
    this.addBytesFromElement(_dataOutputStream, this.classFile);
    return byteArrayOutputStream.toByteArray();
  }
  
  private void addBytesFromElement(final DataOutputStream stream, final EObject element) {
    try {
      if ((element instanceof U1)) {
        stream.writeByte(ClassFileAccessAPI.intValue(((U1)element)));
      } else {
        if ((element instanceof U2)) {
          stream.writeShort(ClassFileAccessAPI.intValue(((U2)element)));
        } else {
          if ((element instanceof U4)) {
            stream.writeInt(ClassFileAccessAPI.intValue(((U4)element)));
          } else {
            if ((element instanceof UString)) {
              stream.writeUTF(((UString)element).getValue());
            } else {
              EList<EStructuralFeature> _eAllStructuralFeatures = element.eClass().getEAllStructuralFeatures();
              for (final EStructuralFeature feature : _eAllStructuralFeatures) {
                if ((feature instanceof EReference)) {
                  boolean _isContainment = ((EReference) feature).isContainment();
                  if (_isContainment) {
                    final Object child = element.eGet(feature);
                    if ((child instanceof EObject)) {
                      this.addBytesFromElement(stream, ((EObject)child));
                    } else {
                      if ((child != null)) {
                        for (final Object c : ((List<?>) child)) {
                          this.addBytesFromElement(stream, ((EObject) c));
                        }
                      }
                    }
                  } else {
                    boolean _isAssignableFrom = ConstantPoolEntry.class.isAssignableFrom(((EReference)feature).getEReferenceType().getInstanceClass());
                    if (_isAssignableFrom) {
                      final Object child_1 = element.eGet(feature);
                      if ((child_1 == null)) {
                        stream.writeByte(0);
                        stream.writeByte(0);
                      } else {
                        if ((child_1 instanceof ConstantPoolEntry)) {
                          final int index = ClassFileAccessAPI.index(((ConstantPoolEntry)child_1));
                          EReference _lDC_Index = JbcPackage.eINSTANCE.getLDC_Index();
                          boolean _tripleEquals = (feature == _lDC_Index);
                          if (_tripleEquals) {
                            stream.writeByte(index);
                          } else {
                            stream.writeShort(index);
                          }
                        }
                      }
                    } else {
                      boolean _isAssignableFrom_1 = CodeTableEntry.class.isAssignableFrom(((EReference)feature).getEReferenceType().getInstanceClass());
                      if (_isAssignableFrom_1) {
                        final Object child_2 = element.eGet(feature);
                        if ((child_2 instanceof CodeTableEntry)) {
                          final int offset = ClassFileAccessAPI.offset(((CodeTableEntry)child_2));
                          if (((((((((((((((((((((feature == JbcPackage.eINSTANCE.getGOTO_Branch()) || 
                            (feature == JbcPackage.eINSTANCE.getGOTO_W_Branch())) || 
                            (feature == JbcPackage.eINSTANCE.getIF_ACMPEQ_Branch())) || 
                            (feature == JbcPackage.eINSTANCE.getIF_ACMPNE_Branch())) || 
                            (feature == JbcPackage.eINSTANCE.getIF_ICMPEQ_Branch())) || 
                            (feature == JbcPackage.eINSTANCE.getIF_ICMPNE_Branch())) || 
                            (feature == JbcPackage.eINSTANCE.getIF_ICMPLT_Branch())) || 
                            (feature == JbcPackage.eINSTANCE.getIF_ICMPGE_Branch())) || 
                            (feature == JbcPackage.eINSTANCE.getIF_ICMPGT_Branch())) || 
                            (feature == JbcPackage.eINSTANCE.getIF_ICMPLE_Branch())) || 
                            (feature == JbcPackage.eINSTANCE.getIFEQ_Branch())) || 
                            (feature == JbcPackage.eINSTANCE.getIFNE_Branch())) || 
                            (feature == JbcPackage.eINSTANCE.getIFLT_Branch())) || 
                            (feature == JbcPackage.eINSTANCE.getIFGE_Branch())) || 
                            (feature == JbcPackage.eINSTANCE.getIFGT_Branch())) || 
                            (feature == JbcPackage.eINSTANCE.getIFLE_Branch())) || 
                            (feature == JbcPackage.eINSTANCE.getIFNONNULL_Branch())) || 
                            (feature == JbcPackage.eINSTANCE.getIFNULL_Branch())) || 
                            (feature == JbcPackage.eINSTANCE.getJSR_W_Branch())) || 
                            (feature == JbcPackage.eINSTANCE.getJSR_Branch()))) {
                            int _offset = ClassFileAccessAPI.offset(((CodeTableEntry) element));
                            int _minus = (offset - _offset);
                            stream.writeShort(_minus);
                          } else {
                            stream.writeShort(offset);
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    } catch (Throwable _e) {
      throw Exceptions.sneakyThrow(_e);
    }
  }
}
