package com.itemis.jbc.binary;

import com.itemis.jbc.binary.Opcode;
import com.itemis.jbc.jbc.AttributeInfo;
import com.itemis.jbc.jbc.Attributes;
import com.itemis.jbc.jbc.ClassFile;
import com.itemis.jbc.jbc.Code;
import com.itemis.jbc.jbc.CodeTable;
import com.itemis.jbc.jbc.CodeTableEntry;
import com.itemis.jbc.jbc.ConstantClass;
import com.itemis.jbc.jbc.ConstantDouble;
import com.itemis.jbc.jbc.ConstantFieldRef;
import com.itemis.jbc.jbc.ConstantFloat;
import com.itemis.jbc.jbc.ConstantInteger;
import com.itemis.jbc.jbc.ConstantInterfaceMethodRef;
import com.itemis.jbc.jbc.ConstantInvoceDynamic;
import com.itemis.jbc.jbc.ConstantLong;
import com.itemis.jbc.jbc.ConstantMethodHandle;
import com.itemis.jbc.jbc.ConstantMethodRef;
import com.itemis.jbc.jbc.ConstantMethodType;
import com.itemis.jbc.jbc.ConstantModule;
import com.itemis.jbc.jbc.ConstantNameAndType;
import com.itemis.jbc.jbc.ConstantPackage;
import com.itemis.jbc.jbc.ConstantPool;
import com.itemis.jbc.jbc.ConstantPoolEntry;
import com.itemis.jbc.jbc.ConstantString;
import com.itemis.jbc.jbc.ConstantUtf8;
import com.itemis.jbc.jbc.ConstantValue;
import com.itemis.jbc.jbc.EnclosingMethod;
import com.itemis.jbc.jbc.Exceptions;
import com.itemis.jbc.jbc.Exports;
import com.itemis.jbc.jbc.FieldInfo;
import com.itemis.jbc.jbc.InnerClasses;
import com.itemis.jbc.jbc.LineNumberTable;
import com.itemis.jbc.jbc.LocalVariableTable;
import com.itemis.jbc.jbc.MethodInfo;
import com.itemis.jbc.jbc.Module;
import com.itemis.jbc.jbc.Opens;
import com.itemis.jbc.jbc.Provides;
import com.itemis.jbc.jbc.SourceFile;
import com.itemis.jbc.jbc.U1;
import com.itemis.jbc.jbc.U2;
import com.itemis.jbc.jbc.U4;
import com.itemis.jbc.jbc.Unknown;
import java.nio.ByteBuffer;
import java.util.HashMap;
import java.util.Map;
import java.util.WeakHashMap;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.xtext.xbase.lib.Conversions;
import org.eclipse.xtext.xbase.lib.Functions.Function1;
import org.eclipse.xtext.xbase.lib.Functions.Function2;
import org.eclipse.xtext.xbase.lib.IterableExtensions;
import org.eclipse.xtext.xbase.lib.ListExtensions;

@SuppressWarnings("all")
public class ClassFileAccessAPI {
  /**
   * Cache for index of constant pool entries inside the constant pool
   */
  private static Map<ConstantPool, Map<ConstantPoolEntry, Integer>> constantPoolIndexCache = new WeakHashMap<ConstantPool, Map<ConstantPoolEntry, Integer>>();
  
  /**
   * Cache for index of code table entries inside each code table
   */
  private static Map<CodeTable, Map<CodeTableEntry, Integer>> codeTableOffsetCache = new WeakHashMap<CodeTable, Map<CodeTableEntry, Integer>>();
  
  private ClassFileAccessAPI() {
  }
  
  public static ClassFile getClassFile(final EObject object) {
    if ((object instanceof ClassFile)) {
      return ((ClassFile) object);
    }
    return ClassFileAccessAPI.getClassFile(object.eContainer());
  }
  
  public static boolean isPublic(final ClassFile classFile) {
    return ClassFileAccessAPI.isAccessSet(classFile, 0x0001);
  }
  
  public static boolean isFinal(final ClassFile classFile) {
    return ClassFileAccessAPI.isAccessSet(classFile, 0x0010);
  }
  
  public static boolean isSuper(final ClassFile classFile) {
    return ClassFileAccessAPI.isAccessSet(classFile, 0x0020);
  }
  
  public static boolean isInterface(final ClassFile classFile) {
    return ClassFileAccessAPI.isAccessSet(classFile, 0x0200);
  }
  
  public static boolean isAbstract(final ClassFile classFile) {
    return ClassFileAccessAPI.isAccessSet(classFile, 0x0400);
  }
  
  public static boolean isSynthetic(final ClassFile classFile) {
    return ClassFileAccessAPI.isAccessSet(classFile, 0x1000);
  }
  
  public static boolean isAnnotation(final ClassFile classFile) {
    return ClassFileAccessAPI.isAccessSet(classFile, 0x2000);
  }
  
  public static boolean isEnum(final ClassFile classFile) {
    return ClassFileAccessAPI.isAccessSet(classFile, 0x4000);
  }
  
  public static boolean isClass(final ClassFile classFile) {
    return (((!ClassFileAccessAPI.isInterface(classFile)) && (!ClassFileAccessAPI.isEnum(classFile))) && (!ClassFileAccessAPI.isAnnotation(classFile)));
  }
  
  public static boolean isPublic(final MethodInfo method) {
    return ClassFileAccessAPI.isAccessSet(method, 0x0001);
  }
  
  public static boolean isPrivate(final MethodInfo method) {
    return ClassFileAccessAPI.isAccessSet(method, 0x0002);
  }
  
  public static boolean isProtected(final MethodInfo method) {
    return ClassFileAccessAPI.isAccessSet(method, 0x0004);
  }
  
  public static boolean isPackageProtected(final MethodInfo method) {
    return (((!ClassFileAccessAPI.isPublic(method)) && (!ClassFileAccessAPI.isProtected(method))) && (!ClassFileAccessAPI.isPrivate(method)));
  }
  
  public static boolean isStatic(final MethodInfo method) {
    return ClassFileAccessAPI.isAccessSet(method, 0x0008);
  }
  
  public static boolean isFinal(final MethodInfo method) {
    return ClassFileAccessAPI.isAccessSet(method, 0x0010);
  }
  
  public static boolean isSynchronized(final MethodInfo method) {
    return ClassFileAccessAPI.isAccessSet(method, 0x0020);
  }
  
  public static boolean isBridge(final MethodInfo method) {
    return ClassFileAccessAPI.isAccessSet(method, 0x0040);
  }
  
  public static boolean isVarargs(final MethodInfo method) {
    return ClassFileAccessAPI.isAccessSet(method, 0x0080);
  }
  
  public static boolean isNative(final MethodInfo method) {
    return ClassFileAccessAPI.isAccessSet(method, 0x0100);
  }
  
  public static boolean isAbstract(final MethodInfo method) {
    return ClassFileAccessAPI.isAccessSet(method, 0x0400);
  }
  
  public static boolean isStrict(final MethodInfo method) {
    return ClassFileAccessAPI.isAccessSet(method, 0x0800);
  }
  
  public static boolean isSynthetic(final MethodInfo method) {
    return ClassFileAccessAPI.isAccessSet(method, 0x1000);
  }
  
  public static boolean isPublic(final FieldInfo field) {
    return ClassFileAccessAPI.isAccessSet(field, 0x0001);
  }
  
  public static boolean isPrivate(final FieldInfo field) {
    return ClassFileAccessAPI.isAccessSet(field, 0x0002);
  }
  
  public static boolean isProtected(final FieldInfo field) {
    return ClassFileAccessAPI.isAccessSet(field, 0x0004);
  }
  
  public static boolean isPackageProtected(final FieldInfo field) {
    return (((!ClassFileAccessAPI.isPublic(field)) && (!ClassFileAccessAPI.isPrivate(field))) && (!ClassFileAccessAPI.isProtected(field)));
  }
  
  public static boolean isStatic(final FieldInfo field) {
    return ClassFileAccessAPI.isAccessSet(field, 0x0008);
  }
  
  public static boolean isFinal(final FieldInfo field) {
    return ClassFileAccessAPI.isAccessSet(field, 0x0010);
  }
  
  public static boolean isVolatile(final FieldInfo field) {
    return ClassFileAccessAPI.isAccessSet(field, 0x0040);
  }
  
  public static boolean isTransient(final FieldInfo field) {
    return ClassFileAccessAPI.isAccessSet(field, 0x0080);
  }
  
  public static boolean isSynthetic(final FieldInfo field) {
    return ClassFileAccessAPI.isAccessSet(field, 0x1000);
  }
  
  public static boolean isEnum(final FieldInfo field) {
    return ClassFileAccessAPI.isAccessSet(field, 0x4000);
  }
  
  private static boolean isAccessSet(final ClassFile classFile, final int mask) {
    U2 _accessFlags = null;
    if (classFile!=null) {
      _accessFlags=classFile.getAccessFlags();
    }
    int _intValue = 0;
    if (_accessFlags!=null) {
      _intValue=ClassFileAccessAPI.intValue(_accessFlags);
    }
    int _bitwiseAnd = (_intValue & mask);
    return (_bitwiseAnd != 0);
  }
  
  private static boolean isAccessSet(final MethodInfo method, final int mask) {
    U2 _accessFlags = null;
    if (method!=null) {
      _accessFlags=method.getAccessFlags();
    }
    int _intValue = 0;
    if (_accessFlags!=null) {
      _intValue=ClassFileAccessAPI.intValue(_accessFlags);
    }
    int _bitwiseAnd = (_intValue & mask);
    return (_bitwiseAnd != 0);
  }
  
  private static boolean isAccessSet(final FieldInfo field, final int mask) {
    U2 _accessFlags = null;
    if (field!=null) {
      _accessFlags=field.getAccessFlags();
    }
    int _intValue = 0;
    if (_accessFlags!=null) {
      _intValue=ClassFileAccessAPI.intValue(_accessFlags);
    }
    int _bitwiseAnd = (_intValue & mask);
    return (_bitwiseAnd != 0);
  }
  
  public static ConstantUtf8 getConstantUtf8(final ClassFile classFile, final int index) {
    return ClassFileAccessAPI.getConstantUtf8(classFile.getConstantPool(), index);
  }
  
  public static ConstantUtf8 getConstantUtf8(final ConstantPool constantPool, final int index) {
    ConstantPoolEntry _constant = null;
    if (constantPool!=null) {
      _constant=ClassFileAccessAPI.getConstant(constantPool, index);
    }
    final ConstantPoolEntry constant = _constant;
    ConstantUtf8 _xifexpression = null;
    if ((constant instanceof ConstantUtf8)) {
      _xifexpression = ((ConstantUtf8)constant);
    } else {
      _xifexpression = null;
    }
    return _xifexpression;
  }
  
  public static ConstantClass getConstantClass(final ConstantPool constantPool, final int index) {
    ConstantPoolEntry _constant = null;
    if (constantPool!=null) {
      _constant=ClassFileAccessAPI.getConstant(constantPool, index);
    }
    final ConstantPoolEntry constant = _constant;
    ConstantClass _xifexpression = null;
    if ((constant instanceof ConstantClass)) {
      _xifexpression = ((ConstantClass)constant);
    } else {
      _xifexpression = null;
    }
    return _xifexpression;
  }
  
  public static ConstantFieldRef getConstantFieldRef(final ConstantPool constantPool, final int index) {
    ConstantPoolEntry _constant = null;
    if (constantPool!=null) {
      _constant=ClassFileAccessAPI.getConstant(constantPool, index);
    }
    final ConstantPoolEntry constant = _constant;
    ConstantFieldRef _xifexpression = null;
    if ((constant instanceof ConstantFieldRef)) {
      _xifexpression = ((ConstantFieldRef)constant);
    } else {
      _xifexpression = null;
    }
    return _xifexpression;
  }
  
  public static ConstantMethodRef getConstantMethodRef(final ConstantPool constantPool, final int index) {
    ConstantPoolEntry _constant = null;
    if (constantPool!=null) {
      _constant=ClassFileAccessAPI.getConstant(constantPool, index);
    }
    final ConstantPoolEntry constant = _constant;
    ConstantMethodRef _xifexpression = null;
    if ((constant instanceof ConstantMethodRef)) {
      _xifexpression = ((ConstantMethodRef)constant);
    } else {
      _xifexpression = null;
    }
    return _xifexpression;
  }
  
  public static ConstantNameAndType getConstantNameAndType(final ConstantPool constantPool, final int index) {
    ConstantPoolEntry _constant = null;
    if (constantPool!=null) {
      _constant=ClassFileAccessAPI.getConstant(constantPool, index);
    }
    final ConstantPoolEntry constant = _constant;
    ConstantNameAndType _xifexpression = null;
    if ((constant instanceof ConstantNameAndType)) {
      _xifexpression = ((ConstantNameAndType)constant);
    } else {
      _xifexpression = null;
    }
    return _xifexpression;
  }
  
  public static ConstantModule getConstantModule(final ConstantPool constantPool, final int index) {
    ConstantPoolEntry _constant = null;
    if (constantPool!=null) {
      _constant=ClassFileAccessAPI.getConstant(constantPool, index);
    }
    final ConstantPoolEntry constant = _constant;
    ConstantModule _xifexpression = null;
    if ((constant instanceof ConstantModule)) {
      _xifexpression = ((ConstantModule)constant);
    } else {
      _xifexpression = null;
    }
    return _xifexpression;
  }
  
  public static ConstantPackage getConstantPackage(final ConstantPool constantPool, final int index) {
    ConstantPoolEntry _constant = null;
    if (constantPool!=null) {
      _constant=ClassFileAccessAPI.getConstant(constantPool, index);
    }
    final ConstantPoolEntry constant = _constant;
    ConstantPackage _xifexpression = null;
    if ((constant instanceof ConstantPackage)) {
      _xifexpression = ((ConstantPackage)constant);
    } else {
      _xifexpression = null;
    }
    return _xifexpression;
  }
  
  public static ConstantPoolEntry getConstant(final ConstantPool constantPool, final int index) {
    ConstantPoolEntry _xblockexpression = null;
    {
      final int realIndex = ClassFileAccessAPI.calculateListIndex(constantPool, index);
      if ((realIndex == (-1))) {
        return null;
      }
      _xblockexpression = constantPool.getCpInfo().get(realIndex);
    }
    return _xblockexpression;
  }
  
  private static int calculateListIndex(final ConstantPool constantPool, final int referenceIndex) {
    int result = 0;
    for (int n = 0; (n < constantPool.getCpInfo().size()); n++) {
      {
        result++;
        final ConstantPoolEntry entry = constantPool.getCpInfo().get(n);
        if ((result == referenceIndex)) {
          return n;
        }
        if (((entry instanceof ConstantLong) || (entry instanceof ConstantDouble))) {
          result++;
        }
      }
    }
    return (-1);
  }
  
  public static int cpSize(final ConstantPool constantPool) {
    int result = 1;
    EList<ConstantPoolEntry> _cpInfo = constantPool.getCpInfo();
    for (final ConstantPoolEntry entry : _cpInfo) {
      {
        result++;
        if (((entry instanceof ConstantLong) || (entry instanceof ConstantDouble))) {
          result++;
        }
      }
    }
    return result;
  }
  
  public static String getStringValue(final ConstantUtf8 constantUtf8) {
    return constantUtf8.getContent().getValue();
  }
  
  public static int getIntValue(final ConstantInteger constantInteger) {
    return ClassFileAccessAPI.intValue(constantInteger.getBytes());
  }
  
  public static float getFloatValue(final ConstantFloat constantFloat) {
    return Float.intBitsToFloat(ClassFileAccessAPI.intValue(constantFloat.getBytes()));
  }
  
  public static long getLongValue(final ConstantLong constantLong) {
    final int highBytes = ClassFileAccessAPI.intValue(constantLong.getHighBytes());
    final int lowBytes = ClassFileAccessAPI.intValue(constantLong.getLowBytes());
    final ByteBuffer buffer = ByteBuffer.allocate(Long.BYTES);
    buffer.putInt(highBytes);
    buffer.putInt(lowBytes);
    buffer.flip();
    return buffer.getLong();
  }
  
  public static double getDoubleValue(final ConstantDouble constantDouble) {
    final int highBytes = ClassFileAccessAPI.intValue(constantDouble.getHighBytes());
    final int lowBytes = ClassFileAccessAPI.intValue(constantDouble.getLowBytes());
    final ByteBuffer buffer = ByteBuffer.allocate(Long.BYTES);
    buffer.putInt(highBytes);
    buffer.putInt(lowBytes);
    buffer.flip();
    return buffer.getDouble();
  }
  
  public static ConstantUtf8 getReferenceNameIndex(final ConstantClass constantClass) {
    return constantClass.getNameIndex();
  }
  
  public static ConstantClass getReferenceClassIndex(final ConstantFieldRef constantFieldRef) {
    return constantFieldRef.getClassIndex();
  }
  
  public static ConstantNameAndType getReferenceNameAndTypeIndex(final ConstantFieldRef constantFieldRef) {
    return constantFieldRef.getNameAndTypeIndex();
  }
  
  public static ConstantClass getReferenceClassIndex(final ConstantMethodRef constantMethodRef) {
    return constantMethodRef.getClassIndex();
  }
  
  public static ConstantNameAndType getReferenceNameAndTypeIndex(final ConstantMethodRef constantMethodRef) {
    return constantMethodRef.getNameAndTypeIndex();
  }
  
  public static ConstantClass getReferenceClassIndex(final ConstantInterfaceMethodRef constantInterfaceMethodRef) {
    return constantInterfaceMethodRef.getClassIndex();
  }
  
  public static ConstantNameAndType getReferenceNameAndTypeIndex(final ConstantInterfaceMethodRef constantInterfaceMethodRef) {
    return constantInterfaceMethodRef.getNameAndTypeIndex();
  }
  
  public static ConstantUtf8 getReferenceStringIndex(final ConstantString constantString) {
    return constantString.getStringIndex();
  }
  
  public static ConstantUtf8 getReferenceNameIndex(final ConstantNameAndType constantNameAndType) {
    return constantNameAndType.getNameIndex();
  }
  
  public static ConstantUtf8 getReferenceDescriptorIndex(final ConstantNameAndType constantNameAndType) {
    return constantNameAndType.getDescriptorIndex();
  }
  
  private static ConstantPool getConstantPool(final ConstantPoolEntry constant) {
    EObject _eContainer = null;
    if (constant!=null) {
      _eContainer=constant.eContainer();
    }
    return ((ConstantPool) _eContainer);
  }
  
  public static int index(final ConstantPoolEntry entry) {
    ConstantPool _constantPool = null;
    if (entry!=null) {
      _constantPool=ClassFileAccessAPI.getConstantPool(entry);
    }
    int _calculateReferenceIndex = 0;
    if (_constantPool!=null) {
      _calculateReferenceIndex=ClassFileAccessAPI.calculateReferenceIndex(_constantPool, entry);
    }
    return _calculateReferenceIndex;
  }
  
  private static int calculateReferenceIndex(final ConstantPool constantPool, final ConstantPoolEntry entry) {
    Map<ConstantPoolEntry, Integer> list = ClassFileAccessAPI.constantPoolIndexCache.get(constantPool);
    if ((list == null)) {
      ClassFileAccessAPI.constantPoolIndexCache.put(constantPool, list = ClassFileAccessAPI.calculateReferenceIndexList(constantPool));
    }
    return (list.get(entry)).intValue();
  }
  
  private static Map<ConstantPoolEntry, Integer> calculateReferenceIndexList(final ConstantPool constantPool) {
    Map<ConstantPoolEntry, Integer> result = new HashMap<ConstantPoolEntry, Integer>();
    int index = 1;
    EList<ConstantPoolEntry> _cpInfo = constantPool.getCpInfo();
    for (final ConstantPoolEntry constant : _cpInfo) {
      {
        result.put(constant, Integer.valueOf(index));
        if (((constant instanceof ConstantDouble) || (constant instanceof ConstantLong))) {
          index++;
        }
        index++;
      }
    }
    return result;
  }
  
  public static Class<? extends ConstantPoolEntry> entryTypeForTag(final int tag) {
    Class<? extends ConstantPoolEntry> _switchResult = null;
    switch (tag) {
      case 1:
        _switchResult = ConstantUtf8.class;
        break;
      case 3:
        _switchResult = ConstantInteger.class;
        break;
      case 4:
        _switchResult = ConstantFloat.class;
        break;
      case 5:
        _switchResult = ConstantLong.class;
        break;
      case 6:
        _switchResult = ConstantDouble.class;
        break;
      case 7:
        _switchResult = ConstantClass.class;
        break;
      case 8:
        _switchResult = ConstantString.class;
        break;
      case 9:
        _switchResult = ConstantFieldRef.class;
        break;
      case 10:
        _switchResult = ConstantMethodRef.class;
        break;
      case 11:
        _switchResult = ConstantInterfaceMethodRef.class;
        break;
      case 12:
        _switchResult = ConstantNameAndType.class;
        break;
      case 15:
        _switchResult = ConstantMethodHandle.class;
        break;
      case 16:
        _switchResult = ConstantMethodType.class;
        break;
      case 18:
        _switchResult = ConstantInvoceDynamic.class;
        break;
      case 19:
        _switchResult = ConstantModule.class;
        break;
      case 20:
        _switchResult = ConstantPackage.class;
        break;
      default:
        throw new RuntimeException(("Unknown tag: " + Integer.valueOf(tag)));
    }
    return _switchResult;
  }
  
  public static int constantTagValue(final ConstantPoolEntry entry) {
    int _switchResult = (int) 0;
    boolean _matched = false;
    if (entry instanceof ConstantUtf8) {
      _matched=true;
      _switchResult = 1;
    }
    if (!_matched) {
      if (entry instanceof ConstantInteger) {
        _matched=true;
        _switchResult = 3;
      }
    }
    if (!_matched) {
      if (entry instanceof ConstantFloat) {
        _matched=true;
        _switchResult = 4;
      }
    }
    if (!_matched) {
      if (entry instanceof ConstantLong) {
        _matched=true;
        _switchResult = 5;
      }
    }
    if (!_matched) {
      if (entry instanceof ConstantDouble) {
        _matched=true;
        _switchResult = 6;
      }
    }
    if (!_matched) {
      if (entry instanceof ConstantClass) {
        _matched=true;
        _switchResult = 7;
      }
    }
    if (!_matched) {
      if (entry instanceof ConstantString) {
        _matched=true;
        _switchResult = 8;
      }
    }
    if (!_matched) {
      if (entry instanceof ConstantFieldRef) {
        _matched=true;
        _switchResult = 9;
      }
    }
    if (!_matched) {
      if (entry instanceof ConstantMethodRef) {
        _matched=true;
        _switchResult = 10;
      }
    }
    if (!_matched) {
      if (entry instanceof ConstantInterfaceMethodRef) {
        _matched=true;
        _switchResult = 11;
      }
    }
    if (!_matched) {
      if (entry instanceof ConstantNameAndType) {
        _matched=true;
        _switchResult = 12;
      }
    }
    if (!_matched) {
      if (entry instanceof ConstantMethodHandle) {
        _matched=true;
        _switchResult = 15;
      }
    }
    if (!_matched) {
      if (entry instanceof ConstantMethodType) {
        _matched=true;
        _switchResult = 16;
      }
    }
    if (!_matched) {
      if (entry instanceof ConstantInvoceDynamic) {
        _matched=true;
        _switchResult = 18;
      }
    }
    if (!_matched) {
      if (entry instanceof ConstantModule) {
        _matched=true;
        _switchResult = 19;
      }
    }
    if (!_matched) {
      if (entry instanceof ConstantPackage) {
        _matched=true;
        _switchResult = 20;
      }
    }
    if (!_matched) {
      throw new RuntimeException(("Unknown tag: " + entry));
    }
    return _switchResult;
  }
  
  public static int offset(final CodeTableEntry entry) {
    CodeTable _table = null;
    if (entry!=null) {
      _table=ClassFileAccessAPI.getTable(entry);
    }
    return (ClassFileAccessAPI.offset(_table, entry)).intValue();
  }
  
  private static Integer offset(final CodeTable table, final CodeTableEntry entry) {
    Map<CodeTableEntry, Integer> offsetTable = ClassFileAccessAPI.codeTableOffsetCache.get(table);
    if ((offsetTable == null)) {
      ClassFileAccessAPI.codeTableOffsetCache.put(table, offsetTable = ClassFileAccessAPI.calculateOffsetTable(table));
    }
    return offsetTable.get(entry);
  }
  
  private static HashMap<CodeTableEntry, Integer> calculateOffsetTable(final CodeTable table) {
    final HashMap<CodeTableEntry, Integer> result = new HashMap<CodeTableEntry, Integer>();
    int offset = 0;
    EList<CodeTableEntry> _instruction = table.getInstruction();
    for (final CodeTableEntry entry : _instruction) {
      {
        result.put(entry, Integer.valueOf(offset));
        int _offset = offset;
        int _byteCount = Opcode.byteCount(entry);
        offset = (_offset + _byteCount);
      }
    }
    return result;
  }
  
  public static CodeTable getTable(final CodeTableEntry entry) {
    EObject _eContainer = entry.eContainer();
    return ((CodeTable) _eContainer);
  }
  
  public static CodeTableEntry instructionFromOffset(final CodeTable table, final int targetOffset) {
    CodeTableEntry actual = null;
    int i = 0;
    int offset = 0;
    do {
      {
        actual = table.getInstruction().get(i);
        if ((offset == targetOffset)) {
          return actual;
        }
        int _offset = offset;
        int _byteCount = Opcode.byteCount(actual);
        offset = (_offset + _byteCount);
        i++;
      }
    } while((offset <= targetOffset));
    return null;
  }
  
  public static int byteCount(final AttributeInfo attribute) {
    boolean _matched = false;
    if (attribute instanceof Code) {
      _matched=true;
      int _byteCount = ClassFileAccessAPI.byteCount(((Code)attribute).getCodeTable());
      int _plus = (((2 + 2) + 4) + _byteCount);
      int _plus_1 = (_plus + 2);
      int _length = ((Object[])Conversions.unwrapArray(((Code)attribute).getExceptionTable().getExceptionTableEntry(), Object.class)).length;
      int _multiply = (_length * 8);
      int _plus_2 = (_plus_1 + _multiply);
      int _plus_3 = (_plus_2 + 2);
      int _byteCount_1 = ClassFileAccessAPI.byteCount(((Code)attribute).getAttributes());
      return (_plus_3 + _byteCount_1);
    }
    if (!_matched) {
      if (attribute instanceof ConstantValue) {
        _matched=true;
        return 2;
      }
    }
    if (!_matched) {
      if (attribute instanceof SourceFile) {
        _matched=true;
        return 2;
      }
    }
    if (!_matched) {
      if (attribute instanceof LineNumberTable) {
        _matched=true;
        int _length = ((Object[])Conversions.unwrapArray(((LineNumberTable)attribute).getLineNumbers(), Object.class)).length;
        int _multiply = (_length * 4);
        return (2 + _multiply);
      }
    }
    if (!_matched) {
      if (attribute instanceof LocalVariableTable) {
        _matched=true;
        int _length = ((Object[])Conversions.unwrapArray(((LocalVariableTable)attribute).getLocalVariables(), Object.class)).length;
        int _multiply = (_length * 10);
        return (2 + _multiply);
      }
    }
    if (!_matched) {
      if (attribute instanceof Exceptions) {
        _matched=true;
        int _length = ((Object[])Conversions.unwrapArray(((Exceptions)attribute).getException(), Object.class)).length;
        int _multiply = (_length * 2);
        return (2 + _multiply);
      }
    }
    if (!_matched) {
      if (attribute instanceof InnerClasses) {
        _matched=true;
        int _length = ((Object[])Conversions.unwrapArray(((InnerClasses)attribute).getInnerClasses(), Object.class)).length;
        int _multiply = (_length * 8);
        return (2 + _multiply);
      }
    }
    if (!_matched) {
      if (attribute instanceof EnclosingMethod) {
        _matched=true;
        return 4;
      }
    }
    if (!_matched) {
      if (attribute instanceof Module) {
        _matched=true;
        int _length = ((Object[])Conversions.unwrapArray(((Module)attribute).getRequires(), Object.class)).length;
        int _multiply = (_length * 6);
        int _plus = ((((2 + 2) + 2) + 2) + _multiply);
        int _plus_1 = (_plus + 2);
        final Function1<Exports, Integer> _function = (Exports e) -> {
          int _length_1 = ((Object[])Conversions.unwrapArray(e.getExportsTo(), Object.class)).length;
          int _multiply_1 = (_length_1 * 2);
          return Integer.valueOf((((2 + 2) + 2) + _multiply_1));
        };
        final Function2<Integer, Integer, Integer> _function_1 = (Integer a, Integer b) -> {
          return Integer.valueOf(((a).intValue() + (b).intValue()));
        };
        Integer _reduce = IterableExtensions.<Integer>reduce(ListExtensions.<Exports, Integer>map(((Module)attribute).getExports(), _function), _function_1);
        int _plus_2 = (_plus_1 + (_reduce).intValue());
        int _plus_3 = (_plus_2 + 2);
        final Function1<Opens, Integer> _function_2 = (Opens e) -> {
          int _length_1 = ((Object[])Conversions.unwrapArray(e.getOpensTo(), Object.class)).length;
          int _multiply_1 = (_length_1 * 2);
          return Integer.valueOf((((2 + 2) + 2) + _multiply_1));
        };
        final Function2<Integer, Integer, Integer> _function_3 = (Integer a, Integer b) -> {
          return Integer.valueOf(((a).intValue() + (b).intValue()));
        };
        Integer _reduce_1 = IterableExtensions.<Integer>reduce(ListExtensions.<Opens, Integer>map(((Module)attribute).getOpens(), _function_2), _function_3);
        int _plus_4 = (_plus_3 + (_reduce_1).intValue());
        int _plus_5 = (_plus_4 + 2);
        int _length_1 = ((Object[])Conversions.unwrapArray(((Module)attribute).getUses(), Object.class)).length;
        int _multiply_1 = (_length_1 * 2);
        int _plus_6 = (_plus_5 + _multiply_1);
        int _plus_7 = (_plus_6 + 2);
        final Function1<Provides, Integer> _function_4 = (Provides e) -> {
          int _length_2 = ((Object[])Conversions.unwrapArray(e.getProvidesWith(), Object.class)).length;
          int _multiply_2 = (_length_2 * 2);
          return Integer.valueOf(((2 + 2) + _multiply_2));
        };
        final Function2<Integer, Integer, Integer> _function_5 = (Integer a, Integer b) -> {
          return Integer.valueOf(((a).intValue() + (b).intValue()));
        };
        Integer _reduce_2 = IterableExtensions.<Integer>reduce(ListExtensions.<Provides, Integer>map(((Module)attribute).getProvides(), _function_4), _function_5);
        return (_plus_7 + (_reduce_2).intValue());
      }
    }
    if (!_matched) {
      if (attribute instanceof Unknown) {
        _matched=true;
        return ((Object[])Conversions.unwrapArray(((Unknown)attribute).getInfo(), Object.class)).length;
      }
    }
    throw new RuntimeException(("Unknown attribute " + attribute));
  }
  
  public static int byteCount(final Attributes attributes) {
    int result = 0;
    EList<AttributeInfo> _attributeInfo = attributes.getAttributeInfo();
    for (final AttributeInfo info : _attributeInfo) {
      int _result = result;
      int _byteCount = ClassFileAccessAPI.byteCount(info);
      int _plus = ((2 + 4) + _byteCount);
      result = (_result + _plus);
    }
    return result;
  }
  
  public static int byteCount(final CodeTable table) {
    int realLength = 0;
    EList<CodeTableEntry> _instruction = table.getInstruction();
    for (final CodeTableEntry entry : _instruction) {
      int _realLength = realLength;
      int _byteCount = Opcode.byteCount(entry);
      realLength = (_realLength + _byteCount);
    }
    return realLength;
  }
  
  public static byte byteValue(final String value) {
    int _parseUnsignedInt = Integer.parseUnsignedInt(value, 16);
    return ((byte) _parseUnsignedInt);
  }
  
  public static int intValue(final U1 u1) {
    int _xblockexpression = (int) 0;
    {
      String _value = u1.getValue();
      boolean _tripleEquals = (_value == null);
      if (_tripleEquals) {
        return 0;
      }
      _xblockexpression = Integer.parseUnsignedInt(u1.getValue(), 16);
    }
    return _xblockexpression;
  }
  
  public static int intValue(final U2 u2) {
    int _xblockexpression = (int) 0;
    {
      String _value = u2.getValue();
      boolean _tripleEquals = (_value == null);
      if (_tripleEquals) {
        return 0;
      }
      _xblockexpression = Integer.parseUnsignedInt(u2.getValue(), 16);
    }
    return _xblockexpression;
  }
  
  public static int intValue(final U4 u4) {
    int _xblockexpression = (int) 0;
    {
      String _value = u4.getValue();
      boolean _tripleEquals = (_value == null);
      if (_tripleEquals) {
        return 0;
      }
      _xblockexpression = Integer.parseUnsignedInt(u4.getValue(), 16);
    }
    return _xblockexpression;
  }
  
  public static String u1Value(final byte value) {
    String result = Integer.toHexString(value).toUpperCase();
    while ((result.length() < 2)) {
      result = ("0" + result);
    }
    return result;
  }
  
  public static String u1Value(final int value) {
    String result = Integer.toHexString(value).toUpperCase();
    while ((result.length() < 2)) {
      result = ("0" + result);
    }
    return result;
  }
  
  public static String u2Value(final int value) {
    String result = Integer.toHexString(value).toUpperCase();
    while ((result.length() < 4)) {
      result = ("0" + result);
    }
    return result;
  }
  
  public static String u2SignedValue(final int value) {
    int unsignedValue = value;
    if ((unsignedValue < 0)) {
      unsignedValue = ((-value) + Short.MAX_VALUE);
    }
    String result = Integer.toHexString(Integer.parseInt(Long.toUnsignedString(unsignedValue, 16), 16)).toUpperCase();
    while ((result.length() < 4)) {
      result = ("0" + result);
    }
    return result;
  }
  
  public static String u4SignedValue(final int value) {
    int unsignedValue = value;
    if ((unsignedValue < 0)) {
      unsignedValue = ((-value) + Integer.MAX_VALUE);
    }
    String result = Integer.toHexString(Integer.parseInt(Long.toUnsignedString(unsignedValue, 16), 16)).toUpperCase();
    while ((result.length() < 8)) {
      result = ("0" + result);
    }
    return result;
  }
  
  public static String u4Value(final int value) {
    String result = Integer.toHexString(value).toUpperCase();
    while ((result.length() < 8)) {
      result = ("0" + result);
    }
    return result;
  }
}
