package com.itemis.jbc.binary;

import com.itemis.jbc.binary.ClassFileAccessAPI;
import com.itemis.jbc.jbc.AALOAD;
import com.itemis.jbc.jbc.AttributeInfo;
import com.itemis.jbc.jbc.Attributes;
import com.itemis.jbc.jbc.ClassFile;
import com.itemis.jbc.jbc.Code;
import com.itemis.jbc.jbc.CodeTable;
import com.itemis.jbc.jbc.CodeTableEntry;
import com.itemis.jbc.jbc.ConstantClass;
import com.itemis.jbc.jbc.ConstantDouble;
import com.itemis.jbc.jbc.ConstantFieldRef;
import com.itemis.jbc.jbc.ConstantFloat;
import com.itemis.jbc.jbc.ConstantInteger;
import com.itemis.jbc.jbc.ConstantInterfaceMethodRef;
import com.itemis.jbc.jbc.ConstantInvoceDynamic;
import com.itemis.jbc.jbc.ConstantLong;
import com.itemis.jbc.jbc.ConstantMethodHandle;
import com.itemis.jbc.jbc.ConstantMethodRef;
import com.itemis.jbc.jbc.ConstantMethodType;
import com.itemis.jbc.jbc.ConstantModule;
import com.itemis.jbc.jbc.ConstantNameAndType;
import com.itemis.jbc.jbc.ConstantPackage;
import com.itemis.jbc.jbc.ConstantPool;
import com.itemis.jbc.jbc.ConstantPoolEntry;
import com.itemis.jbc.jbc.ConstantString;
import com.itemis.jbc.jbc.ConstantUtf8;
import com.itemis.jbc.jbc.ConstantValue;
import com.itemis.jbc.jbc.EnclosingMethod;
import com.itemis.jbc.jbc.ExceptionTable;
import com.itemis.jbc.jbc.ExceptionTableEntry;
import com.itemis.jbc.jbc.Exports;
import com.itemis.jbc.jbc.FieldInfo;
import com.itemis.jbc.jbc.Fields;
import com.itemis.jbc.jbc.GOTO;
import com.itemis.jbc.jbc.Interface;
import com.itemis.jbc.jbc.Interfaces;
import com.itemis.jbc.jbc.JbcFactory;
import com.itemis.jbc.jbc.JumpOffset;
import com.itemis.jbc.jbc.LDC;
import com.itemis.jbc.jbc.LOOKUPSWITCH;
import com.itemis.jbc.jbc.LineNumber;
import com.itemis.jbc.jbc.LineNumberTable;
import com.itemis.jbc.jbc.MatchOffsetPair;
import com.itemis.jbc.jbc.MethodInfo;
import com.itemis.jbc.jbc.Methods;
import com.itemis.jbc.jbc.Module;
import com.itemis.jbc.jbc.NOP;
import com.itemis.jbc.jbc.Opens;
import com.itemis.jbc.jbc.Provides;
import com.itemis.jbc.jbc.Requires;
import com.itemis.jbc.jbc.SourceFile;
import com.itemis.jbc.jbc.TABLESWITCH;
import com.itemis.jbc.jbc.U1;
import com.itemis.jbc.jbc.U2;
import com.itemis.jbc.jbc.U4;
import com.itemis.jbc.jbc.UString;
import com.itemis.jbc.jbc.Unknown;
import com.itemis.jbc.jbc.Uses;
import org.eclipse.xtext.xbase.lib.CollectionExtensions;

@SuppressWarnings("all")
public class ClassFileFactoryAPI {
  private ClassFileFactoryAPI() {
  }
  
  public static ClassFile classFile(final U4 magic, final U2 minorVersion, final U2 majorVersion, final U2 constantPoolCount, final ConstantPool constantPool, final U2 accessFlags, final ConstantClass thisClass, final ConstantClass superClass, final U2 interfacesCount, final Interfaces interfaces, final U2 fieldsCount, final Fields fields, final U2 methodsCount, final Methods methods, final U2 attributesCount, final Attributes attributes) {
    final ClassFile result = JbcFactory.eINSTANCE.createClassFile();
    result.setMagic(magic);
    result.setMinorVersion(minorVersion);
    result.setMajorVersion(majorVersion);
    result.setConstantPoolCount(constantPoolCount);
    result.setConstantPool(constantPool);
    result.setAccessFlags(accessFlags);
    result.setThisClass(thisClass);
    result.setSuperClass(superClass);
    result.setInterfaceCount(interfacesCount);
    result.setInterfaces(interfaces);
    result.setFieldsCount(fieldsCount);
    result.setFields(fields);
    result.setMethodsCount(methodsCount);
    result.setMethods(methods);
    result.setAttributesCount(attributesCount);
    result.setAttributes(attributes);
    return result;
  }
  
  public static ConstantPool constantPool(final ConstantPoolEntry... entries) {
    final ConstantPool result = JbcFactory.eINSTANCE.createConstantPool();
    for (final ConstantPoolEntry entry : entries) {
      result.getCpInfo().add(entry);
    }
    return result;
  }
  
  public static ConstantUtf8 constantUtf8(final U1 tag, final UString content) {
    final ConstantUtf8 result = JbcFactory.eINSTANCE.createConstantUtf8();
    result.setTag(tag);
    result.setContent(content);
    return result;
  }
  
  public static ConstantInteger constantInteger(final U1 tag, final U4 bytes) {
    final ConstantInteger result = JbcFactory.eINSTANCE.createConstantInteger();
    result.setTag(tag);
    result.setBytes(bytes);
    return result;
  }
  
  public static ConstantFloat constantFloat(final U1 tag, final U4 bytes) {
    final ConstantFloat result = JbcFactory.eINSTANCE.createConstantFloat();
    result.setTag(tag);
    result.setBytes(bytes);
    return result;
  }
  
  public static ConstantLong constantLong(final U1 tag, final U4 highBytes, final U4 lowBytes) {
    final ConstantLong result = JbcFactory.eINSTANCE.createConstantLong();
    result.setTag(tag);
    result.setHighBytes(highBytes);
    result.setLowBytes(lowBytes);
    return result;
  }
  
  public static ConstantDouble constantDouble(final U1 tag, final U4 highBytes, final U4 lowBytes) {
    final ConstantDouble result = JbcFactory.eINSTANCE.createConstantDouble();
    result.setTag(tag);
    result.setHighBytes(highBytes);
    result.setLowBytes(lowBytes);
    return result;
  }
  
  public static ConstantClass constantClass(final U1 tag, final ConstantUtf8 nameIndex) {
    final ConstantClass result = JbcFactory.eINSTANCE.createConstantClass();
    result.setTag(tag);
    result.setNameIndex(nameIndex);
    return result;
  }
  
  public static ConstantString constantString(final U1 tag, final ConstantUtf8 stringIndex) {
    final ConstantString result = JbcFactory.eINSTANCE.createConstantString();
    result.setTag(tag);
    result.setStringIndex(stringIndex);
    return result;
  }
  
  public static ConstantFieldRef constantFieldRef(final U1 tag, final ConstantClass classIndex, final ConstantNameAndType nameAndTypeIndex) {
    final ConstantFieldRef result = JbcFactory.eINSTANCE.createConstantFieldRef();
    result.setTag(tag);
    result.setClassIndex(classIndex);
    result.setNameAndTypeIndex(nameAndTypeIndex);
    return result;
  }
  
  public static ConstantMethodRef constantMethodRef(final U1 tag, final ConstantClass classIndex, final ConstantNameAndType nameAndTypeIndex) {
    final ConstantMethodRef result = JbcFactory.eINSTANCE.createConstantMethodRef();
    result.setTag(tag);
    result.setClassIndex(classIndex);
    result.setNameAndTypeIndex(nameAndTypeIndex);
    return result;
  }
  
  public static ConstantInterfaceMethodRef constantInterfaceMethodRef(final U1 tag, final ConstantClass classIndex, final ConstantNameAndType nameAndTypeIndex) {
    final ConstantInterfaceMethodRef result = JbcFactory.eINSTANCE.createConstantInterfaceMethodRef();
    result.setTag(tag);
    result.setClassIndex(classIndex);
    result.setNameAndTypeIndex(nameAndTypeIndex);
    return result;
  }
  
  public static ConstantNameAndType constantNameAndType(final U1 tag, final ConstantUtf8 nameIndex, final ConstantUtf8 descriptorIndex) {
    final ConstantNameAndType result = JbcFactory.eINSTANCE.createConstantNameAndType();
    result.setTag(tag);
    result.setNameIndex(nameIndex);
    result.setDescriptorIndex(descriptorIndex);
    return result;
  }
  
  public static ConstantMethodHandle constantMethodHandle(final U1 tag, final U1 referenceKind, final ConstantNameAndType referenceIndex) {
    final ConstantMethodHandle result = JbcFactory.eINSTANCE.createConstantMethodHandle();
    result.setTag(tag);
    result.setReferenceKind(referenceKind);
    result.setReferenceIndex(referenceIndex);
    return result;
  }
  
  public static ConstantMethodType constantMethodType(final U1 tag, final ConstantUtf8 descriptorIndex) {
    final ConstantMethodType result = JbcFactory.eINSTANCE.createConstantMethodType();
    result.setTag(tag);
    result.setDescriptorIndex(descriptorIndex);
    return result;
  }
  
  public static ConstantInvoceDynamic constantInvoceDynamic(final U1 tag, final U2 bootstrapMethodAttrIndex, final ConstantNameAndType nameAndTypeIndex) {
    final ConstantInvoceDynamic result = JbcFactory.eINSTANCE.createConstantInvoceDynamic();
    result.setTag(tag);
    result.setBootstrapMethodAttrIndex(bootstrapMethodAttrIndex);
    result.setNameAndTypeIndex(nameAndTypeIndex);
    return result;
  }
  
  public static ConstantModule constantModule(final U1 tag, final ConstantUtf8 nameIndex) {
    final ConstantModule result = JbcFactory.eINSTANCE.createConstantModule();
    result.setTag(tag);
    result.setNameIndex(nameIndex);
    return result;
  }
  
  public static ConstantPackage constantPackage(final U1 tag, final ConstantUtf8 nameIndex) {
    final ConstantPackage result = JbcFactory.eINSTANCE.createConstantPackage();
    result.setTag(tag);
    result.setNameIndex(nameIndex);
    return result;
  }
  
  public static Interfaces interfaces(final Interface... interfaceInfo) {
    final Interfaces result = JbcFactory.eINSTANCE.createInterfaces();
    CollectionExtensions.<Interface>addAll(result.getInterfaceInfo(), interfaceInfo);
    return result;
  }
  
  public static Interface interfaceInfo(final ConstantClass info) {
    final Interface result = JbcFactory.eINSTANCE.createInterface();
    result.setInfo(info);
    return result;
  }
  
  public static Fields fields(final FieldInfo... fieldInfo) {
    final Fields result = JbcFactory.eINSTANCE.createFields();
    CollectionExtensions.<FieldInfo>addAll(result.getFieldInfo(), fieldInfo);
    return result;
  }
  
  public static FieldInfo fieldInfo(final U2 accessFlags, final ConstantUtf8 nameIndex, final ConstantUtf8 descriptorIndex, final U2 attributesCount, final Attributes attributes) {
    final FieldInfo result = JbcFactory.eINSTANCE.createFieldInfo();
    result.setAccessFlags(accessFlags);
    result.setNameIndex(nameIndex);
    result.setDescriptorIndex(descriptorIndex);
    result.setAttributesCount(attributesCount);
    result.setAttributes(attributes);
    return result;
  }
  
  public static Methods methods(final MethodInfo... methodInfo) {
    final Methods result = JbcFactory.eINSTANCE.createMethods();
    CollectionExtensions.<MethodInfo>addAll(result.getMethodsInfo(), methodInfo);
    return result;
  }
  
  public static MethodInfo methodInfo(final U2 accessFlags, final ConstantUtf8 nameIndex, final ConstantUtf8 descriptorIndex, final U2 attributesCount, final Attributes attributes) {
    final MethodInfo result = JbcFactory.eINSTANCE.createMethodInfo();
    result.setAccessFlags(accessFlags);
    result.setNameIndex(nameIndex);
    result.setDescriptorIndex(descriptorIndex);
    result.setAttributesCount(attributesCount);
    result.setAttributes(attributes);
    return result;
  }
  
  public static Attributes attributes(final AttributeInfo... attributeInfo) {
    final Attributes result = JbcFactory.eINSTANCE.createAttributes();
    CollectionExtensions.<AttributeInfo>addAll(result.getAttributeInfo(), attributeInfo);
    return result;
  }
  
  public static Unknown attributeUnknown(final ConstantUtf8 attributeNameIndex, final U4 attributeLength, final U1... info) {
    final Unknown result = JbcFactory.eINSTANCE.createUnknown();
    result.setAttributeNameIndex(attributeNameIndex);
    result.setAttributeLength(attributeLength);
    for (final U1 i : info) {
      result.getInfo().add(i);
    }
    return result;
  }
  
  public static ConstantValue attributeConstantValue(final ConstantUtf8 attributeNameIndex, final U4 attributeLength, final ConstantPoolEntry constantValueIndex) {
    final ConstantValue result = JbcFactory.eINSTANCE.createConstantValue();
    result.setAttributeNameIndex(attributeNameIndex);
    result.setAttributeLength(attributeLength);
    result.setConstantValueIndex(constantValueIndex);
    return result;
  }
  
  public static Code attributeCode(final ConstantUtf8 attributeNameIndex, final U4 attributeLength, final U2 maxStack, final U2 maxLocals, final U4 codeLength, final CodeTable codeTable, final U2 exceptionTableLength, final ExceptionTable exceptionTable, final U2 attributesCount, final Attributes attributes) {
    final Code result = JbcFactory.eINSTANCE.createCode();
    result.setAttributeNameIndex(attributeNameIndex);
    result.setAttributeLength(attributeLength);
    result.setMaxStack(maxStack);
    result.setMaxLocals(maxLocals);
    result.setCodeLength(codeLength);
    result.setCodeTable(codeTable);
    result.setExceptionTableLength(exceptionTableLength);
    result.setExceptionTable(exceptionTable);
    result.setAttributesCount(attributesCount);
    result.setAttributes(attributes);
    return result;
  }
  
  public static EnclosingMethod attributeEnclosingMethod(final ConstantUtf8 attributeNameIndex, final U4 attributeLength, final ConstantClass classIndex, final ConstantNameAndType methodIndex) {
    final EnclosingMethod result = JbcFactory.eINSTANCE.createEnclosingMethod();
    result.setAttributeNameIndex(attributeNameIndex);
    result.setAttributeLength(attributeLength);
    result.setClassIndex(classIndex);
    result.setMethodIndex(methodIndex);
    return result;
  }
  
  public static ExceptionTable exceptionTable(final ExceptionTableEntry... exceptionTableEntry) {
    final ExceptionTable result = JbcFactory.eINSTANCE.createExceptionTable();
    CollectionExtensions.<ExceptionTableEntry>addAll(result.getExceptionTableEntry(), exceptionTableEntry);
    return result;
  }
  
  public static ExceptionTableEntry exceptionTableEntry(final CodeTableEntry startPc, final CodeTableEntry endPc, final CodeTableEntry handlerPc, final ConstantClass catchType) {
    final ExceptionTableEntry result = JbcFactory.eINSTANCE.createExceptionTableEntry();
    result.setStartPc(startPc);
    result.setEndPc(endPc);
    result.setHandlerPc(handlerPc);
    result.setCatchType(catchType);
    return result;
  }
  
  public static CodeTable codeTable(final CodeTableEntry... copdeTableEntry) {
    final CodeTable result = JbcFactory.eINSTANCE.createCodeTable();
    CollectionExtensions.<CodeTableEntry>addAll(result.getInstruction(), copdeTableEntry);
    return result;
  }
  
  public static SourceFile attributeSourceFile(final ConstantUtf8 attributeNameIndex, final U4 attributeLength, final ConstantUtf8 sourceFileIndex) {
    final SourceFile result = JbcFactory.eINSTANCE.createSourceFile();
    result.setAttributeNameIndex(attributeNameIndex);
    result.setAttributeLength(attributeLength);
    result.setSourceFileIndex(sourceFileIndex);
    return result;
  }
  
  public static LineNumberTable attributeLineNumberTable(final ConstantUtf8 attributeNameIndex, final U4 attributeLength, final U2 lineNumberTableLength, final LineNumber... lines) {
    final LineNumberTable result = JbcFactory.eINSTANCE.createLineNumberTable();
    result.setAttributeNameIndex(attributeNameIndex);
    result.setAttributeLength(attributeLength);
    result.setLineNumberTableLength(lineNumberTableLength);
    CollectionExtensions.<LineNumber>addAll(result.getLineNumbers(), lines);
    return result;
  }
  
  public static LineNumber lineNumber(final CodeTableEntry startPc, final U2 lineNumber) {
    final LineNumber result = JbcFactory.eINSTANCE.createLineNumber();
    result.setStartPc(startPc);
    result.setLineNumber(lineNumber);
    return result;
  }
  
  public static Module attributeModule(final ConstantUtf8 attributeNameIndex, final U4 attributeLength, final ConstantModule moduleNameIndex, final U2 moduleFlags, final ConstantUtf8 moduleVersionIndex, final U2 requiresCount, final Requires[] requires, final U2 exportsCount, final Exports[] exports, final U2 opensCount, final Opens[] opens, final U2 usesCount, final Uses[] uses, final U2 providesCount, final Provides[] provides) {
    final Module result = JbcFactory.eINSTANCE.createModule();
    result.setAttributeNameIndex(attributeNameIndex);
    result.setAttributeLength(attributeLength);
    result.setModuleNameIndex(moduleNameIndex);
    result.setModuleFlags(moduleFlags);
    result.setModuleVersionIndex(moduleVersionIndex);
    result.setRequiresCount(requiresCount);
    CollectionExtensions.<Requires>addAll(result.getRequires(), requires);
    result.setExportsCount(exportsCount);
    CollectionExtensions.<Exports>addAll(result.getExports(), exports);
    result.setOpensCount(opensCount);
    CollectionExtensions.<Opens>addAll(result.getOpens(), opens);
    result.setUsesCount(usesCount);
    CollectionExtensions.<Uses>addAll(result.getUses(), uses);
    result.setProvidesCount(providesCount);
    CollectionExtensions.<Provides>addAll(result.getProvides(), provides);
    return result;
  }
  
  public static U1 u1(final int value) {
    final U1 result = JbcFactory.eINSTANCE.createU1();
    result.setValue(ClassFileAccessAPI.u1Value(value));
    return result;
  }
  
  public static U2 u2(final int value) {
    final U2 result = JbcFactory.eINSTANCE.createU2();
    result.setValue(ClassFileAccessAPI.u2Value(value));
    return result;
  }
  
  public static U4 u4(final int value) {
    final U4 result = JbcFactory.eINSTANCE.createU4();
    result.setValue(ClassFileAccessAPI.u4Value(value));
    return result;
  }
  
  public static UString uString(final String value) {
    final UString result = JbcFactory.eINSTANCE.createUString();
    result.setValue(value);
    return result;
  }
  
  public static AALOAD aaload(final U1 tag) {
    final AALOAD result = JbcFactory.eINSTANCE.createAALOAD();
    result.setTag(tag);
    return result;
  }
  
  public static LDC ldc(final U1 tag, final ConstantPoolEntry index) {
    final LDC result = JbcFactory.eINSTANCE.createLDC();
    result.setTag(tag);
    result.setIndex(index);
    return result;
  }
  
  public static GOTO goTo(final U1 tag, final CodeTableEntry branch) {
    final GOTO result = JbcFactory.eINSTANCE.createGOTO();
    result.setTag(tag);
    result.setBranch(branch);
    return result;
  }
  
  public static LOOKUPSWITCH lookupswitch(final U1 tag, final U4 default_, final U4 npairs, final MatchOffsetPair... matchOffsetPairs) {
    final LOOKUPSWITCH result = JbcFactory.eINSTANCE.createLOOKUPSWITCH();
    result.setTag(tag);
    result.setDefault(default_);
    result.setNpairs(npairs);
    CollectionExtensions.<MatchOffsetPair>addAll(result.getMatchOffsetPairs(), matchOffsetPairs);
    return result;
  }
  
  public static LOOKUPSWITCH lookupswitch(final U1 tag, final U1 pad0, final U4 default_, final U4 npairs, final MatchOffsetPair... matchOffsetPairs) {
    final LOOKUPSWITCH result = JbcFactory.eINSTANCE.createLOOKUPSWITCH();
    result.setTag(tag);
    result.setPad0(pad0);
    result.setDefault(default_);
    result.setNpairs(npairs);
    CollectionExtensions.<MatchOffsetPair>addAll(result.getMatchOffsetPairs(), matchOffsetPairs);
    return result;
  }
  
  public static LOOKUPSWITCH lookupswitch(final U1 tag, final U1 pad0, final U1 pad1, final U4 default_, final U4 npairs, final MatchOffsetPair... matchOffsetPairs) {
    final LOOKUPSWITCH result = JbcFactory.eINSTANCE.createLOOKUPSWITCH();
    result.setTag(tag);
    result.setPad0(pad0);
    result.setPad1(pad1);
    result.setDefault(default_);
    result.setNpairs(npairs);
    CollectionExtensions.<MatchOffsetPair>addAll(result.getMatchOffsetPairs(), matchOffsetPairs);
    return result;
  }
  
  public static LOOKUPSWITCH lookupswitch(final U1 tag, final U1 pad0, final U1 pad1, final U1 pad2, final U4 default_, final U4 npairs, final MatchOffsetPair... matchOffsetPairs) {
    final LOOKUPSWITCH result = JbcFactory.eINSTANCE.createLOOKUPSWITCH();
    result.setTag(tag);
    result.setPad0(pad0);
    result.setPad1(pad1);
    result.setPad2(pad2);
    result.setDefault(default_);
    result.setNpairs(npairs);
    CollectionExtensions.<MatchOffsetPair>addAll(result.getMatchOffsetPairs(), matchOffsetPairs);
    return result;
  }
  
  public static MatchOffsetPair matchOffsetPair(final U4 match, final U4 offset) {
    final MatchOffsetPair result = JbcFactory.eINSTANCE.createMatchOffsetPair();
    result.setMatch(match);
    result.setOffset(offset);
    return result;
  }
  
  public static TABLESWITCH tableswitch(final U1 tag, final U4 default_, final U4 low, final U4 high, final JumpOffset... jumpOffsets) {
    final TABLESWITCH result = JbcFactory.eINSTANCE.createTABLESWITCH();
    result.setTag(tag);
    result.setDefault(default_);
    result.setLow(low);
    result.setHigh(high);
    CollectionExtensions.<JumpOffset>addAll(result.getJumpOffsets(), jumpOffsets);
    return result;
  }
  
  public static TABLESWITCH tableswitch(final U1 tag, final U1 pad0, final U4 default_, final U4 low, final U4 high, final JumpOffset... jumpOffsets) {
    final TABLESWITCH result = JbcFactory.eINSTANCE.createTABLESWITCH();
    result.setTag(tag);
    result.setDefault(default_);
    result.setPad0(pad0);
    result.setLow(low);
    result.setHigh(high);
    CollectionExtensions.<JumpOffset>addAll(result.getJumpOffsets(), jumpOffsets);
    return result;
  }
  
  public static TABLESWITCH tableswitch(final U1 tag, final U1 pad0, final U1 pad1, final U4 default_, final U4 low, final U4 high, final JumpOffset... jumpOffsets) {
    final TABLESWITCH result = JbcFactory.eINSTANCE.createTABLESWITCH();
    result.setTag(tag);
    result.setDefault(default_);
    result.setPad0(pad0);
    result.setPad1(pad1);
    result.setLow(low);
    result.setHigh(high);
    CollectionExtensions.<JumpOffset>addAll(result.getJumpOffsets(), jumpOffsets);
    return result;
  }
  
  public static TABLESWITCH tableswitch(final U1 tag, final U1 pad0, final U1 pad1, final U1 pad2, final U4 default_, final U4 low, final U4 high, final JumpOffset... jumpOffsets) {
    final TABLESWITCH result = JbcFactory.eINSTANCE.createTABLESWITCH();
    result.setTag(tag);
    result.setDefault(default_);
    result.setPad0(pad0);
    result.setPad1(pad1);
    result.setPad2(pad2);
    result.setLow(low);
    result.setHigh(high);
    CollectionExtensions.<JumpOffset>addAll(result.getJumpOffsets(), jumpOffsets);
    return result;
  }
  
  public static JumpOffset jumpOffset(final U4 offset) {
    final JumpOffset result = JbcFactory.eINSTANCE.createJumpOffset();
    result.setOffset(offset);
    return result;
  }
  
  public static NOP nop(final U1 tag) {
    final NOP result = JbcFactory.eINSTANCE.createNOP();
    result.setTag(tag);
    return result;
  }
}
