/**
 * generated by Xtext 2.11.0
 */
package com.itemis.jbc.validation;

import com.itemis.jbc.binary.ClassFileAccessAPI;
import com.itemis.jbc.binary.Opcode;
import com.itemis.jbc.jbc.AttributeInfo;
import com.itemis.jbc.jbc.ClassFile;
import com.itemis.jbc.jbc.Code;
import com.itemis.jbc.jbc.CodeTableEntry;
import com.itemis.jbc.jbc.ConstantPoolEntry;
import com.itemis.jbc.jbc.FieldInfo;
import com.itemis.jbc.jbc.LineNumberTable;
import com.itemis.jbc.jbc.LocalVariableTable;
import com.itemis.jbc.jbc.MethodInfo;
import com.itemis.jbc.jbc.U1;
import com.itemis.jbc.jbc.U2;
import com.itemis.jbc.jbc.U4;
import com.itemis.jbc.validation.AbstractJBCValidator;
import org.eclipse.xtext.validation.Check;
import org.eclipse.xtext.xbase.lib.Conversions;

/**
 * This class contains custom validation rules.
 * 
 * See https://www.eclipse.org/Xtext/documentation/303_runtime_concepts.html#validation
 */
@SuppressWarnings("all")
public class JBCValidator extends AbstractJBCValidator {
  public final static String EXPECTED_VALUE = "expectedValue";
  
  @Check
  public void classFileAttributes(final ClassFile classFile) {
    this.expect(classFile.getMagic(), (-889275714));
    this.expect(classFile.getConstantPoolCount(), ClassFileAccessAPI.cpSize(classFile.getConstantPool()));
    this.expect(classFile.getInterfaceCount(), ((Object[])Conversions.unwrapArray(classFile.getInterfaces().getInterfaceInfo(), Object.class)).length);
    this.expect(classFile.getMethodsCount(), ((Object[])Conversions.unwrapArray(classFile.getMethods().getMethodsInfo(), Object.class)).length);
    this.expect(classFile.getFieldsCount(), ((Object[])Conversions.unwrapArray(classFile.getFields().getFieldInfo(), Object.class)).length);
    this.expect(classFile.getAttributesCount(), ((Object[])Conversions.unwrapArray(classFile.getAttributes().getAttributeInfo(), Object.class)).length);
  }
  
  @Check
  public void constantPoolEntryAttributes(final ConstantPoolEntry entry) {
    this.expect(entry.getTag(), ClassFileAccessAPI.constantTagValue(entry));
  }
  
  @Check
  public void fieldInfoAttributes(final FieldInfo fieldInfo) {
    this.expect(fieldInfo.getAttributesCount(), ((Object[])Conversions.unwrapArray(fieldInfo.getAttributes().getAttributeInfo(), Object.class)).length);
  }
  
  @Check
  public void methodInfoAttributes(final MethodInfo methodInfo) {
    this.expect(methodInfo.getAttributesCount(), ((Object[])Conversions.unwrapArray(methodInfo.getAttributes().getAttributeInfo(), Object.class)).length);
  }
  
  @Check
  public void attributeInfoAttributes(final AttributeInfo attribute) {
    this.expect(attribute.getAttributeLength(), ClassFileAccessAPI.byteCount(attribute));
  }
  
  @Check
  public void codeAttributes(final Code codeAttribute) {
    this.expect(codeAttribute.getCodeLength(), ClassFileAccessAPI.byteCount(codeAttribute.getCodeTable()));
    this.expect(codeAttribute.getExceptionTableLength(), ((Object[])Conversions.unwrapArray(codeAttribute.getExceptionTable().getExceptionTableEntry(), Object.class)).length);
    this.expect(codeAttribute.getAttributesCount(), ((Object[])Conversions.unwrapArray(codeAttribute.getAttributes().getAttributeInfo(), Object.class)).length);
  }
  
  @Check
  public void codeTableEntryAttributes(final CodeTableEntry entry) {
    this.expect(entry.getTag(), Opcode.from(entry).code);
  }
  
  @Check
  public void lineNumberTableAttributes(final LineNumberTable table) {
    this.expect(table.getLineNumberTableLength(), ((Object[])Conversions.unwrapArray(table.getLineNumbers(), Object.class)).length);
  }
  
  @Check
  public void localVariableTableAttributes(final LocalVariableTable table) {
    this.expect(table.getLocalVariableTableLength(), ((Object[])Conversions.unwrapArray(table.getLocalVariables(), Object.class)).length);
  }
  
  private void expect(final U1 attribute, final int expectedValue) {
    int _intValue = ClassFileAccessAPI.intValue(attribute);
    boolean _notEquals = (_intValue != expectedValue);
    if (_notEquals) {
      String _u1Value = ClassFileAccessAPI.u1Value(expectedValue);
      String _plus = ("Expected \'" + _u1Value);
      String _plus_1 = (_plus + "\'");
      this.error(_plus_1, attribute.eContainingFeature(), JBCValidator.EXPECTED_VALUE, ClassFileAccessAPI.u1Value(expectedValue));
    }
  }
  
  private void expect(final U2 attribute, final int expectedValue) {
    int _intValue = ClassFileAccessAPI.intValue(attribute);
    boolean _notEquals = (_intValue != expectedValue);
    if (_notEquals) {
      String _u2Value = ClassFileAccessAPI.u2Value(expectedValue);
      String _plus = ("Expected \'" + _u2Value);
      String _plus_1 = (_plus + "\'");
      this.error(_plus_1, attribute.eContainingFeature(), JBCValidator.EXPECTED_VALUE, ClassFileAccessAPI.u2Value(expectedValue));
    }
  }
  
  private void expect(final U4 attribute, final int expectedValue) {
    int _intValue = ClassFileAccessAPI.intValue(attribute);
    boolean _notEquals = (_intValue != expectedValue);
    if (_notEquals) {
      String _u4Value = ClassFileAccessAPI.u4Value(expectedValue);
      String _plus = ("Expected \'" + _u4Value);
      String _plus_1 = (_plus + "\'");
      this.error(_plus_1, attribute.eContainingFeature(), JBCValidator.EXPECTED_VALUE, ClassFileAccessAPI.u4Value(expectedValue));
    }
  }
}
